/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.commons.events;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.collection.IListEvaluator;


/**
 * Event registry that lets the user register for events they are interested
 * in. When an event has occurred all the registered event handlers are
 * invoked.
 * 
 * @author Oliver Plohmann
 */
public class EventRegistry<T extends EventNotifier> extends EventRegistryCore<T> implements IEventRegistry<T>
{
	/**
	 * Creates a new <code>EventRegistry</code> object.
	 */
	public EventRegistry()
	{
		super();
	}

	/**
	 * Register for an event providing a list of event handlers that should be
	 * invoked on event occurrence.
	 * 
	 * @param eventName
	 * @param notifier
	 * @throws IllegalArgumentException
	 * @see org.almendra.commons.events.IEventRegistry#register(org.almendra.commons.events.Event,
	 *      org.almendra.commons.events.EventNotifier)
	 */
	public void register(String eventName, IListEvaluator<T> list) throws IllegalArgumentException
	{
		if (!StringUtility.isValid(eventName))
			throw new IllegalArgumentException("no name specified for event"); //$NON-NLS-1$
		accessProtection.lock();
		List<T> notifiers = registry.get(eventName);
		if (notifiers == null)
		{
			notifiers = new LinkedList<T>();
			registry.put(eventName, notifiers);
		}
		notifiers.add(getEventNotifier(list));
		accessProtection.unlock();
	}

	/**
	 * @param list
	 * @return
	 */
	protected EventNotifier getEventNotifier(IListEvaluator<T> list)
	{
		return new EventNotifier(list);
	}

	/**
	 * Deregister for event. All event handlers for the event are removed from
	 * the event registry.
	 * 
	 * @param eventName
	 * @return @see org.almendra.commons.commons.events.IEventRegistry#deregister(org.almendra.commons.commons.events.Event)
	 */
	public boolean deregister(String eventName, IListEvaluator<T> list) throws NoSuchElementException
	{
		accessProtection.lock();
		try
		{
			List<T> notifiers = registry.get(eventName);
			if (notifiers == null)
			{	
				throw new NoSuchElementException("no event registered named " + eventName); //$NON-NLS-1$
			}
			List<T> copiedNotifiers = new ArrayList<T>(notifiers);
			for(T notifier : copiedNotifiers)
			{
				if (notifier.getEventHandlers() == list)
				{	
					boolean result = notifiers.remove(notifier);
					return result;
				}
			}
			throw new NoSuchElementException("list evaluator for event named " + eventName + " not found!"); //$NON-NLS-1$ //$NON-NLS-2$
		}
		finally
		{
		    accessProtection.unlock();
		}				
	}

}
