/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.commons.threads;

import org.almendra.commons.INamedObject;
import org.almendra.commons.threads.ISimpleActiveObject;


/**
 * Abstract active object that constantly calls <code>doActivity()</code> as long as
 * it is active. An active object is an object that runs in its own thread. Defines
 * several auxiliary methods to be used by concrete subclasses.
 * 
 * @author Oliver Plohmann
 */
public abstract class ActiveObject implements Runnable, ISimpleActiveObject, INamedObject
{
	/**
	 * Name of the active object. Optionally can be set to identify active objects
	 * when tracing their activities. 
	 */
	protected String name = ""; //$NON-NLS-1$
	
	/**
	 * Specifies whether the active object should write trace information to the console.
	 */
	protected boolean verbose = false;
	
	/**
	 * Thread the active object runs in.
	 */
	protected transient Thread thread = null;
	
	/**
	 * Holds information whether active object is active or not 
	 * (has been started or has been stopped).
	 */
	protected boolean isActive = false;
	
	/**
	 * Lock object used to synchronize access the flag <code>isActive</code>. 
	 */
	protected Object activeLock = new Object();
	
	/**
	 * Defines the priority the active object should have when started. 
	 */
	protected int priority = Thread.NORM_PRIORITY;
	
	/**
	 * Flag that must be set to true when active object is blocked when
	 * it is waiting for input.  
	 */
	protected boolean waiting = false;

	/**
	 * Creates a new <code>ActiveObject</code> object.
	 *  
	 */
	public ActiveObject()
	{
		super();
	}

	/**
	 * Creates a new <code>ActiveObject</code> object.
	 * @param verbose define whether trace information should be written to the console
	 */
	public ActiveObject(boolean verbose)
	{
		super();
		setVerbose(verbose);
	}

	/**
	 * Creates a new <code>ActiveObject</code> object.
	 * @param verbose define whether trace information should be written to the console
	 * @param priority define priority of thread when active object is started.
	 */
	public ActiveObject(boolean verbose, int priority)
	{
		super();
		setVerbose(verbose);
		this.priority = priority;
	}
	
	/**
	 * Constantly calls <code>doActivity</code> as long as active object is active.
	 * @see java.lang.Runnable#run()
	 */
	public void run()
	{
		while (isActive())
		{
			Thread currentThread = Thread.currentThread();
			while (thread == currentThread)
			{
				doActivity();
			}
		}
	}

	/**
	 * Abstract method that is constantly called as long as active object is active.
	 * Must be implemented by every concrete subclass, for example to fetch items from
	 * an input list and process them as long as the list contains items.
	 */
	protected abstract void doActivity();

	/**
	 * Answer the user-defined name of the active object.
	 * @see org.almendra.commons.cas.INamedObject#getName() 
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * Set user-defined name of the active object. Displayed when 
	 * trace informnation is written to the console.
	 * 
	 * @param name
	 */
	public void setName(String name)
	{
		this.name = name;
	}

	/**
	 * @return priority of the thread the active object is running in
	 */
	public int getPriority()
	{
		if (isActive())
			thread.getPriority();
		return priority;
	}

	/**
	 * Set priority of active object. The active object will run in a thread of the
	 * specified object when started or its priority will be changed accordingly
	 * when already started.
	 * @param priority
	 */
	public void setPriority(int priority)
	{
		synchronized (activeLock)
		{
			if (isActive)
				thread.setPriority(priority);
			this.priority = priority;
		}
	}

	/**
	 * Answer whether the active object has been started and is active or not.
	 */
	public boolean isActive()
	{
		synchronized (activeLock)
		{
			return isActive;
		}
	}

	/**
	 * Start the active object. Ignored if already started.
	 */
	public synchronized void start()
	{
		synchronized (activeLock)
		{
			if (!isActive)
			{
				thread = new Thread(this);
				thread.setDaemon(isDaemon());
				thread.setPriority(priority);
				isActive = true;
				if (isVerbose())
					messageStarting();
				thread.start();
			}
		}
	}

	protected void messageStarting()
	{
		System.out.println("started active object " + getName()); //$NON-NLS-1$
	}

	/**
	 * Stop the active object. Ignored if never started or already stopped.
	 */
	public synchronized void stop()
	{
		synchronized (activeLock)
		{
			if (!isActive)
				return;
			setActive(false);
			thread = null;
			notify();
			if (isVerbose())
				messageStopping();
		}
	}
	
	protected void messageStopping()
	{
		System.out.println("active object " + getName() + " stopped"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * Mark the active object as active or not as indicated by <code>isActive</code>.
	 * @param isActive
	 */
	protected void setActive(boolean isActive)
	{
		synchronized (activeLock)
		{
			this.isActive = isActive;
		}
	}

	/**
	 * Implement in subclass whether thread the active object will run in when started
	 * is a daemon thread or not. A daemon thread only lives to server user threads.
	 * 
	 * @return
	 */
	protected abstract boolean isDaemon();

	/**
	 * Answer whether active object is in verbose mode writing trace 
	 * information to the console.
	 * @return
	 */
	protected boolean isVerbose()
	{
		return verbose;
	}

	/**
	 * Set the active object to verbose mode which will cause trace information 
	 * to be written to the console.
	 * @param verbose
	 */
	public void setVerbose(boolean verbose)
	{
		this.verbose = verbose;
	}

}
