/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.commons.threads;

import java.util.Queue;
import java.util.concurrent.Semaphore;

import org.almendra.commons.util.logging.ErrorLogger;


/**
 * Defines an active object that consumes items provided by some producer
 * object and processes them as long as the list of items to be processed is
 * not empty. If it becomes empty the <code>ActiveSchedulingObject</code>
 * falls asleep till a new item has been added to the list.
 * 
 * @author Oliver Plohmann
 */
public abstract class ActiveSchedulingObject<ExecutableType extends IExecutable> extends ActiveObject
{
	/**
	 * Queue that holds the items to be processed.
	 */
	protected Queue<ExecutableType> queue = null;

	/**
	 * Lock used to synchronize access to the <code>queue</code>.
	 */
	protected Semaphore available = new Semaphore(0);
	
	/**
	 * Answer whether is currently executing an object.
	 */
	protected volatile boolean executing = false;

	/**
	 * Creates a new <code>ActiveSchedulingObject</code> object. Declared
	 * protected to enforce the use of the public constructors by objects that
	 * are not subclasses. The public constructors guarantee propper
	 * initialization with default values. Subclasses that want to set other
	 * user-defined values need to make sure themselves that the object is
	 * initialized propperly.
	 */
	protected ActiveSchedulingObject()
	{
		super();
	}

	/**
	 * Creates a new <code>ActiveSchedulingObject</code> object.
	 * 
	 * @param queue
	 *            the queue that holds items to be processed
	 */
	public ActiveSchedulingObject(Queue<ExecutableType> queue)
	{
		super();
		this.queue = queue;
	}

	/**
	 * Creates a new <code>ActiveSchedulingObject</code> object.
	 * 
	 * @param queue
	 *            the queue that holds items to be processed
	 * @param verbose
	 *            defines wether trace information is written to the console
	 */
	public ActiveSchedulingObject(Queue<ExecutableType> queue, boolean verbose)
	{
		super(verbose);
		this.queue = queue;
	}

	/**
	 * Creates a new <code>ActiveSchedulingObject</code> object.
	 * 
	 * @param queue
	 *            the queue that holds items to be processed
	 * @param verbose
	 *            defines wether trace information is written to the console
	 * @param priority
	 *            the priority of the thread the active will run in
	 */
	public ActiveSchedulingObject(Queue<ExecutableType> queue, boolean verbose, int priority)
	{
		super(verbose, priority);
		this.queue = queue;
	}

	/**
	 * Constantly called by the superclass as long as the active object is
	 * active. For class <code>ActiveSchedulingObject</code> items to be
	 * processed are constantly picked from <code>queue</code> as long as
	 * this <code>queue</code> is not empty. Otherwise the active object
	 * falls asleep and wakes up when a new item has been added.
	 * 
	 * @see org.almendra.commons.cas.scheduling.ActiveObject#doActivity()
	 */
	protected void doActivity()
	{
		ExecutableType next = getNext();
		if (next != null)
		{
			executePrimary(next);
		}
	}

	/**
	 * Called to process the next item from the <code>queue</code>. Must be
	 * implemented in subclasses as appropriate.
	 * @param next
	 */
	protected abstract void execute(ExecutableType executable);
	
	/**
	 * Called to process the next item from the <code>queue</code>.
	 * 
	 * @param next
	 *            the next item from the list to be processed
	 */
	protected void executePrimary(ExecutableType executable)
	{
		try
		{
			startingExecution(executable);
			executing = true;
			execute(executable);
			executing = false;
			finishedExecution(executable);			
		}
		catch (Exception ex)
		{
			System.out.println("exception occured executing " + executable + " !"); //$NON-NLS-1$ //$NON-NLS-2$
			System.out.println(ex.getMessage());
			ex.printStackTrace();
			ErrorLogger.exception(ex);
		}
		finally
		{
			executing = false;
			finishedExecution(executable);
		}
	}

	/**
	 * @param next
	 */
	protected void finishedExecution(ExecutableType executable)
	{
		// redefine in subclasses as appropriate
	}

	/**
	 * @param next
	 */
	protected void startingExecution(ExecutableType executable)
	{
		//	redefine in subclasses as appropriate
	}

	/**
	 * Adds another item to the <code>queue</code> of items to be processed.
	 * Calls <code>objectAdded()</code> to notify instances of subclasses
	 * that an item has been added to the item list.
	 * 
	 * @param object
	 */
	public void add(ExecutableType executable)
	{
		queue.add(executable);		
		available.release();
		executableAdded(executable);
	}

	/**
	 * Called whentever a new item has been added to the <code>queue</code>.
	 * May be redefined in subclasses as appropriate.
	 */
	protected void executableAdded(ExecutableType executable)
	{
		// redefine in subclass as appropriate
	}

	/**
	 * Remove an item from the <code>queue</code>.
	 * 
	 * @param object the object to be removed
	 * @return returns the removed object or null if not found
	 */
	public Object remove(ExecutableType executable)
	{
	    available.acquireUninterruptibly();
		boolean changed = queue.remove(executable);
        if(changed)
            executableRemoved(executable);
		return executable;
	}

	/**
	 * 
	 */
	protected void executableRemoved(ExecutableType executable)
	{
		//	redefine in subclass as appropriate
	}

	/**
	 * Return the next item of the <code>queue</code> to be processed. Blocks
	 * if <code>queue</code> is empty.
	 * 
	 * @return returns next item in <code>queue</code> or null but only if
	 *         the active object was stopped
	 */
	protected ExecutableType getNext()
	{
		if (!hasNext())
		{
			suspending();
			ExecutableType next = getNextPrim();
			resuming();			
			return next;
		}
		return getNextPrim();
	}
	
	protected void resuming()
	{
		waiting = false;
		if (isActive() && isVerbose())
			messageResuming();
	}

	/**
	 * 
	 */
	protected void messageResuming()
	{
		System.out.println("active object " + getName() + " wake up"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	protected void suspending()
	{
		if (isVerbose())
			messageSuspending();
		waiting = true;			
	}

	/**
	 * 
	 */
	protected void messageSuspending()
	{
		System.out.println("active object " + getName() + " waiting ..."); //$NON-NLS-1$ //$NON-NLS-2$
	}

	public boolean hasNext()
	{
		return available.availablePermits() > 0;
	}
	
	public long available()
	{
		return available.availablePermits();
	}

	protected ExecutableType getNextPrim()
	{
		if (isActive())
		{
			available.acquireUninterruptibly();
			return queue.poll();
		}
		return null;
	}
	
	public boolean isIdle()
	{
		return available() <= 0 && !executing;
	}
	
	/**
	 * @return
	 */
	public long getQueueSize()
	{
		return available();
	}

}
