/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.commons.threads;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.almendra.commons.IHistorizable;
import org.almendra.commons.IHistoryItem;
import org.almendra.commons.IHistorySummary;
import org.almendra.commons.INamedObject;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.logging.ErrorLogger;

/**
 * @author Oliver
 *
 */
public abstract class Future<T extends IHistoryItem> implements INamedObject, ISignalable, IHistorizable<T>
{
    protected static final int LATCH_COUNT = 1;
    
	protected List<T> historyItems = new ArrayList<T>();
	
	/**
	 * Creates a new <code>Future</code> object.
	 */
	public Future()
	{
		super();
	}
	
	/**
	 * Creates a new <code>Future</code> object.
	 */
	public Future(String name)
	{
		super();
		this.name = name;
	}

	/**
	 * User-defined name of the future. Can be used to write trace information
	 * to the console to determine which future is waiting for a result.
	 */
	protected String name = null;
	
	/**
	 * Used to block the calling thread if it requests the result that has not been
	 * produced so far. A latch is a condition that is acquirable forever more after 
	 * the Nth release.
	 */
	protected CountDownLatch latch = new CountDownLatch(LATCH_COUNT);
	
	/**
	 * Answers whether a timeout occurred waiting for the result.
	 */
	protected boolean timeoutOccurred = false;
	
	/**
	 * Variable set to true when the <code>latch</code> has been released. Used to determine
	 * whether latch has been released explicitly after result has arrived or whether a
	 * timeout has occurred.
	 */
	protected boolean signaled = false;
	
	/**
	 * Used to synchronize access to the <code>signaled</code> and <code>timeoutOccurred</code> state variable.
	 */
	protected Lock accessProtection = new ReentrantLock(true);
	
	/**
	 * Stores exception that may occur waiting for result. 
	 */
	protected Exception exception = null;
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.INamedObject#getName()
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @return Returns the timeout.
	 */
	public boolean isTimeout()
	{
	    accessProtection.lock();
	    try
		{
	        return timeoutOccurred;
		}
	    finally
	    {
	        accessProtection.unlock();
	    }
	}

	/**
	 * Flag that indicates whether trace information should be written to the console
	 * when the future is waiting for the result.
	 */
	protected volatile boolean verbose = false;

	/**
	 * @param verbose The verbose to set.
	 */
	public void setVerbose(boolean verbose)
	{
		this.verbose = verbose;
	}

	/**
	 * @return Returns the verbose.
	 */
	public boolean isVerbose()
	{
		return verbose;
	}

	/**
	 * @return Returns the exception.
	 */
	public Exception getException()
	{
		return exception;
	}

	/**
	 * @param exception The exception to set.
	 */
	public void setException(Exception exception)
	{
		this.exception = exception;
	}

	/**
	 * Make the calling thread wait till the result has arrived. Can be called explicitly without
	 * calling <code>getResult(...)</code> when the calling thread only has to be blocked till some
	 * event has occurred and there is no interest in some result.
	 */
	public void acquire()
	{
		try
		{
			if(isVerbose())
				System.out.println("future " + getName() + " waiting for result ..."); //$NON-NLS-1$ //$NON-NLS-2$
			latch.await();
		}
		catch (InterruptedException e)
		{
		    latch.countDown();
			setException(e);
			System.err.println(e.getMessage());
			e.printStackTrace();
			ErrorLogger.exception(e);
		}
	}

	/**
	 * Make the calling thread wait till the result has arrived. Can be called explicitly without
	 * calling <code>getResult(...)</code> when the calling thread only has to be blocked till some
	 * event has occurred and there is no interest in some result.
	 * @param timeoutInMillisOrZero timeout period in milliseconds after which the calling thread is released
	 * @throws IllegalArgumentException thrown if <code>timeoutInMillisOrZero</code> smaller than zero
	 */
	public void acquire(long timeoutInMillisOrZero)
	throws IllegalArgumentException, TimeoutException
	{
		if(timeoutInMillisOrZero < 0)
			throw new IllegalArgumentException(" timeoutInMillisOrZero must be 0 or greater!"); //$NON-NLS-1$
		if(timeoutInMillisOrZero == 0)
		{
			acquire();
			return;
		}
		
		try
		{
			primAcquire(timeoutInMillisOrZero);
		}
		catch (TimeoutException e)
		{
			setException(e);
			throw e;
		}
	}

	/**
	 * Utility method to block the calling thread until the result has arrived or the timeout period expired. 
	 * @param milliseconds the timeout period in milliseconds
	 * @throws TimeoutException thrown when timeout period expired waiting for the result
	 */
	protected void primAcquire(long milliseconds)
	throws TimeoutException
	{
		try
		{			
			if(isVerbose())
				System.out.println("future " + getName() + " waiting for result ..."); //$NON-NLS-1$ //$NON-NLS-2$
			
			latch.await(milliseconds, TimeUnit.MILLISECONDS);

			accessProtection.lock();
			try
			{			
				if(!signaled)
				{
					timeoutOccurred = true;
					throw new TimeoutException(
						"timeout period expired waiting for latch of future " +	//$NON-NLS-1$ 
						StringUtility.print(getName()));
				}
			}
			finally
			{
			    accessProtection.unlock();
			}
		}
		catch (InterruptedException e)
		{
		    latch.countDown();
			setException(e);
			System.err.println(e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * Anser whether an exception has occurred waiting for the result.
	 * @return
	 */
	public boolean hasException()
	{
		return getException() != null;
	}

	/**
	 * Unblock the calling thread in case the result has arrived or some event has occurred.
	 */
	public void signal()
	{				
	    accessProtection.lock();
	    try
		{			
			signaled = true;
			latch.countDown();
		}
	    finally
	    {
	        accessProtection.unlock();
	    }
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.threads.ISignalable#holds()
	 */
	public boolean holds()
	{
	    return latch.getCount() == LATCH_COUNT;
	}

	/**
	 * Returns null if called before <code>acquire()</code> has resumed.
	 */
	/* (non-Javadoc)
	 * @see org.almendra.commons.IHistorizable#getHistoryItems()
	 */
	public List<T> getHistoryItems()
	{
		return historyItems;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.IHistorizable#createHistorySummary(org.almendra.commons.IHistorySummary)
	 */
	public void createHistorySummary(IHistorySummary<T> historySummary)
	{
		historySummary.create(getHistoryItems());
	}

	public void addHistoryItems(List<T> items)
	{
		historyItems.addAll(items);
	}
	
}
