/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.commons.threads;

import java.io.Serializable;

/**
 * Holds a promise to deliver a result after unlimited time or after a specifiable period.
 * Used to make the calling thread wait until the result has been produced by some other thread.
 * Upon arrival of the result the calling thread is unblocked or when the timeout period has
 * expired. In the latter case null is returned as the result. 
 * 
 * @author Oliver Plohmann
 */

public class FutureResult extends Future
{
	
	/**
	 * Creates a new <code>FutureResult</code> object.
	 */
	public FutureResult()
	{
		super();
	}
	
	/**
	 * 
	 * Creates a new <code>FutureResult</code> object.
	 * @param name the user-defined name of the future
	 */
	public FutureResult(String name)
	{
		super();
		this.name = name;
	}	
	
	/**
	 * Creates a new <code>FutureResult</code> object.
	 * @param name the user-defined name of the future
	 * @param verbose indicates whether trace information should be written to the console
	 */
	public FutureResult(String name, boolean verbose)
	{
		super();
		this.name = name;
		this.verbose = verbose;
	}		

	/**
	 * @return Returns the result.
	 */
	public Serializable getResult()
	{
		if(result == null)
		{
			acquire();
		}
		return result;
	}
	
	/**
	 * Returns the result. Blocks till result has arrived or timeout has occurred.
	 * @param timeoutPeriod timeout period in milliseconds
	 * @return the result
	 */
	public Serializable getResult(int timeoutPeriod) 
	throws IllegalArgumentException, TimeoutException
	{
	    if(result == null)
		{
	        accessProtection.lock();
	        boolean wasUnlocked = false;
	        try
	        {
	            if(timeoutOccurred)
	                return result;
	            wasUnlocked = true;
	            accessProtection.unlock();
	            acquire(timeoutPeriod);
	        }
	        finally
	        {
	            if(!wasUnlocked)
	                accessProtection.unlock();
	        }	        
		}
		return result;
	}

	/**
	 * @param result The result to set.
	 */
	public void setResult(Serializable result)
	{
		this.result = result;
		if(isVerbose())
			System.out.println("future " + getName() + " has received result"); //$NON-NLS-1$ //$NON-NLS-2$
		signal();
	}

	/**
	 * Attribute used to store the result the calling thread is asking for.
	 */
	protected volatile Serializable result = null;

}
