/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.commons.util.inet;

import java.rmi.AccessException;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.collection.StringList;


/**
 * @author Oliver
 *
 */
public class RMIUtility
{
	protected static Map<String, Registry> Registries = new HashMap<String, Registry>();
	
	/**
	 * Creates a new <code>RMIUtility</code> object.
	 */
	public RMIUtility()
	{
		super();
	}
	
	public static Registry getRegistry() throws RemoteException
	{
		return getRegistry(null, Registry.REGISTRY_PORT);
	}
	
	public static Registry getRegistry(String hostName, int port) throws RemoteException
	{
		String key = getKey(hostName, port);
		Registry registry = Registries.get(key);
		
		if(registry == null)
		{	
			int tempPort = Registry.REGISTRY_PORT;
			if(port > 0)
				tempPort = port;
			if(StringUtility.isValid(hostName))
			{				
				registry = LocateRegistry.getRegistry(hostName, tempPort);
			}
			else
			{
				registry = LocateRegistry.getRegistry(tempPort);
			}			
			checkAlive(registry);
			Registries.put(key, registry);
		}
		
		return registry;
	}
	
	/**
	 * @param hostName
	 * @param port
	 * @return
	 */
	private static String getKey(String hostName, int port)
	{
		if(!StringUtility.isValid(hostName))
			return String.valueOf(port);
		return hostName + ":" + port; //$NON-NLS-1$
	}

	public static Registry createRegistry(int port) 
	throws RemoteException
	{
		int tempPort = Registry.REGISTRY_PORT;
		if(port == -1)
			tempPort = port; 
		Registry registry = LocateRegistry.createRegistry(tempPort);
		checkAlive(registry);
		return registry;
	}

	/**
	 * @param registry
	 */
	protected static void checkAlive(Registry registry) 
	throws AccessException, RemoteException
	{
		registry.list();	// check whether stub is dead or alive
	}

	public static void printListToConsole(String hostName, int port) 
	{
		String[] list = null;
		try
		{
			list = getRegistry(hostName, port).list();
			for (int i = 0; i < list.length; i++)
			{
				System.out.println(list[i]);
			}
		}
		catch (AccessException e)
		{
			list = null;
		}
		catch (RemoteException e)
		{
			list = null;
		}
		if(list == null)
			System.out.println("no registry found at " + getKey(hostName, port)); //$NON-NLS-1$

	}

	public static void printListToConsole() 
	{
		printListToConsole(null, Registry.REGISTRY_PORT);
	}
	
	public static void printListToConsole(int port) 
	{
		printListToConsole(null, port);
	}

	/**
	 * @param location
	 * @return
	 */
	public static Registry getRegistry(RegistryLocation location) throws RemoteException
	{
		return getRegistry(location.getHostName(), location.getPort());
	}
	
	public static Registry ensureRegistryExists(RegistryLocation location) 
	throws RemoteException
	{
		Registry registry = null;

		try
		{
			registry = RMIUtility.getRegistry(location);
		}
		catch (RemoteException e)
		{
			registry = null; // no registry found, so a registry has to be created
		}

		if (registry == null)
		{
			registry = RMIUtility.createRegistry(location.getPort());
		}
		
		return registry;
	}
	
	public static List<String> unbind(RegistryLocation registryLocation, String[] registryKeyPrefixes) 
	throws AccessException, RemoteException, NotBoundException
	{
		List<String> unboundObjectsRegistryKeys = new StringList();
		Registry registry = RMIUtility.getRegistry(registryLocation);
		if(registry == null)
			return new StringList();
		
		String[] registryKeys = registry.list();
		for (int i = 0; i < registryKeys.length; i++)
		{
			String registryKey = registryKeys[i];
			for (int j = 0; j < registryKeyPrefixes.length; j++)
			{
				if(registryKey.startsWith(registryKeyPrefixes[j]))
				{	
					registry.unbind(registryKey);
					unboundObjectsRegistryKeys.add(registryKey);
				}
			}				
		}
		return unboundObjectsRegistryKeys;
	}
}
