/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.commons.util.logging;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.TimeUtility;
import org.almendra.commons.util.collection.StringList;
import org.almendra.commons.util.collection.StringSet;

/**
 * @author Oliver Plohmann
 */

public class Logger<E extends LogEntry, L extends Logger> implements ILogPrinter
{
	public static final int DEBUG = 0;
	public static final int ERROR = 1;
	public static final int FATAL = 2;
	public static final int INFO = 3;
	public static final int WARNING = 4;
	public static final int EXCEPTION = 5;
	public static final int FREE_TEXT = 6;

	public static final String NewSession = "// ---- NEW SESSION ----"; //$NON-NLS-1$

	protected Logger parent = null;
	protected long localLogFileMaxSize = -1;
	protected boolean localLogFileMaxSizeChecked = false;
	protected ILogPrinter logPrinter = null;
	protected String[] levelLabels = null;
	protected Set<String> registeredEvents = new StringSet();
	protected Set<String> subscribedEvents = new StringSet();
	protected List<L> children = new LinkedList<L>();
	protected String localLogFileName = null;
	protected FileWriter localLogWriter = null;
	protected boolean bufferUnwrittenMsgs = false;
	protected List<String> logBuffer = new StringList();
	protected List<ILogListener<LogEntry>> listeners = new LinkedList<ILogListener<LogEntry>>();

	/**
	 * Lock access to log writer when it is being changed since there are
	 * several threads that write to the log which could cause race conditions.
	 */
	protected Lock accessProtection = new ReentrantLock();

	public Logger()
	{
		super();
		init();
	}

	public Logger(boolean bufferUnwrittenMsgs, long localLogFileMaxSize)
	{
		super();
		init();
		this.bufferUnwrittenMsgs = bufferUnwrittenMsgs;
		this.localLogFileMaxSize = localLogFileMaxSize;
	}

	public Logger(ILogPrinter logPrinter)
	{
		init();
		this.logPrinter = logPrinter;
	}

	public Logger(ILogPrinter logPrinter, boolean bufferUnwrittenMsgs)
	{
		this(logPrinter);
		this.bufferUnwrittenMsgs = bufferUnwrittenMsgs;
	}

	public void debug(E entry)
	{
		entry.setLevel(DEBUG);
		log(entry);
	}

	public void debug(String logEntryLabel)
	{
		if (isSubscribed(logEntryLabel))
		{
			E entry = getNewLogEntry();
			entry.setEvent(logEntryLabel);
			debug(entry);
		}
	}

	public void exception(LogEntry entry)
	{
		entry.setLevel(EXCEPTION);
		log(entry);
	}

	public void freeText(LogEntry entry)
	{
		entry.setLevel(FREE_TEXT);
		log(entry);
	}
	
	public void freeText(String freeText)
	{
		LogEntry entry = getNewLogEntry();
		entry.setArgument(freeText);
		freeText(entry);
	}
	
	public void info(LogEntry entry)
	{
		entry.setLevel(INFO);
		log(entry);
	}

	public void info(String logEntryLabel)
	{
		if (isSubscribed(logEntryLabel))
		{
			LogEntry entry = getNewLogEntry();
			entry.setEvent(logEntryLabel);
			info(entry);
		}
	}

	public void warn(String logEntryLabel)
	{
		if (isSubscribed(logEntryLabel))
		{
			LogEntry entry = getNewLogEntry();
			entry.setEvent(logEntryLabel);
			warn(entry);
		}
	}

	public void warn(LogEntry entry)
	{
		entry.setLevel(WARNING);
		log(entry);
	}

	public void error(String logEntryLabel)
	{
		if (isSubscribed(logEntryLabel))
		{
			LogEntry entry = getNewLogEntry();
			entry.setEvent(logEntryLabel);
			error(entry);
		}
	}

	public void error(LogEntry entry)
	{
		entry.setLevel(ERROR);
		log(entry);
	}

	public void fatal(String logEntryLabel)
	{
		if (isSubscribed(logEntryLabel))
		{
			LogEntry entry = getNewLogEntry();
			entry.setEvent(logEntryLabel);
			fatal(entry);
		}
	}
	
	public void fatalText(String text)
	{
		LogEntry entry = getNewLogEntry();
		entry.setArgument(text);
		fatal(entry);
	}
	
	

	public void fatal(LogEntry entry)
	{
		entry.setLevel(FATAL);
		log(entry);
	}

	protected void init()
	{
		levelLabels = new String[7];
		levelLabels[DEBUG] = "DEBUG"; //$NON-NLS-1$
		levelLabels[ERROR] = "ERROR"; //$NON-NLS-1$
		levelLabels[FATAL] = "FATAL"; //$NON-NLS-1$
		levelLabels[INFO] = "INFO"; //$NON-NLS-1$
		levelLabels[WARNING] = "WARNING"; //$NON-NLS-1$
		levelLabels[EXCEPTION] = "EXCEPTION"; //$NON-NLS-1$
		levelLabels[FREE_TEXT] = "FREETEXT"; //$NON-NLS-1$

		logPrinter = new SystemLogPrinter();
	}

	protected void log(LogEntry entry)
	{
		if(entry.mustBeSubscribed() && !isSubscribed(entry.getEvent()))
			return;

		if (parent != null)
		{
			parent.primLog(entry);
			return;
		}

		entry.setLogMessage(levelLabels);
		sendToListeners(entry);
		log(entry.getLogMessage());
	}

	/**
	 * @param entry
	 */
	protected void sendToListeners(LogEntry entry)
	{
		AssertUtility.assertTrue(StringUtility.isValid(entry.getLogMessage()));
		accessProtection.lock();
		try
		{
			if (listeners.isEmpty())
				return;
			for(ILogListener<LogEntry> listener : listeners)
				listener.log(entry);    
		}
		finally
		{
		    accessProtection.unlock();
		}		
	}

	protected void primLog(LogEntry entry)
	{
		if (parent != null)
		{
			parent.primLog(entry);
			return;
		}

		entry.setLogMessage(levelLabels);
		sendToListeners(entry);

		String logMsg = entry.getLogMessage();
		logPrinter.log(logMsg);
		logLocalWriter(logMsg);
	}

	public boolean isRegistered(String eventName)
	{
		return registeredEvents.contains(eventName);
	}

	public boolean isSubscribed(String eventName)
	{
		return subscribedEvents.contains(eventName);
	}

	public void subscribeForEvents(List<String> stringsOfEventNames)
	{
		subscribedEvents.addAll(stringsOfEventNames);
	}

	public void unsubscribeForEvents(List<String> stringsOfEventNames)
	{
		subscribedEvents.removeAll(stringsOfEventNames);
	}

	public void registerForEvents(List<String> stringsOfEventNames)
	{
		registeredEvents.addAll(stringsOfEventNames);
	}

	public void deregisterForEvents(List<String> stringsOfEventNames)
	{
		registeredEvents.removeAll(stringsOfEventNames);
	}

	public LoggerAccessor getPublicAccessor()
	{
		return new LoggerAccessor(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.logging.ILogPrinter#log(java.lang.String)
	 */
	public void log(String msg)
	{
		logPrinter.log(msg);
		logLocalWriter(msg);
	}

	protected void logLocalWriter(String msg)
	{
		if (localLogWriter != null)
		{
			accessProtection.lock();
			try
			{
				localLogWriter.write(msg + "\n"); //$NON-NLS-1$
				localLogWriter.flush();
			}
			catch (IOException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
			finally
			{
			    accessProtection.unlock();
			}
		}
		else if (bufferUnwrittenMsgs)
		{
			logBuffer.add(msg);
		}
	}

	/**
	 * Returns a child logger. A child logger inherits all the attribute values
	 * of the parent logger, but new events can be registered without affecting
	 * the parent logger.
	 * 
	 * @param logger
	 *            the parent logger
	 * @return the child logger
	 */
	public static Logger getChildLogger(Logger logger)
	{
		if (logger == null)
			return null;
		Logger childLogger = new Logger();
		return getChildLogger(logger, childLogger);
	}

	/**
	 * Helper method to allow for logger to be filled from subclass with values
	 * defined in this class.
	 * 
	 * @param parentLogger
	 * @param childLogger
	 * @return
	 */
	protected static <L extends Logger> L getChildLogger(L parentLogger, L childLogger)
	{
		childLogger.parent = parentLogger;
		parentLogger.children.add(childLogger);
		childLogger.logPrinter = parentLogger.logPrinter;
		childLogger.registeredEvents = parentLogger.registeredEvents;
		childLogger.subscribedEvents = parentLogger.subscribedEvents;
		return childLogger;
	}

	public void setLocalLogFileName(String filePathName) throws IOException
	{		
		AssertUtility.assertTrue("log file name already set. use <changeLocalLogFileName>", localLogFileName == null); //$NON-NLS-1$
		accessProtection.lock();
		try
		{
			checkLocalLogMaxSize(filePathName);
			localLogWriter = new FileWriter(new File(filePathName), true);
			localLogFileName = filePathName;
			writeBuffer();		
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/**
	 * @param filePathName
	 */
	protected void checkLocalLogMaxSize(String filePathName)
	{
		if (localLogFileMaxSize == -1)
			return;
		if (localLogFileMaxSizeChecked)
			return;
		File file = new File(filePathName);
		if (file.exists())
		{
			if (file.length() > localLogFileMaxSize)
			{
				file.delete();
				localLogFileMaxSizeChecked = true;
			}
		}
	}

	protected void writeBuffer() throws IOException
	{
		if (bufferUnwrittenMsgs)
		{
			accessProtection.lock();
			try
			{
				for(String msg : logBuffer)
					localLogWriter.write(msg + "\n"); //$NON-NLS-1$
				bufferUnwrittenMsgs = false;
				logBuffer = new StringList();			    
			}
			finally
			{
			    accessProtection.unlock();
			}			
		}
	}

	public void addListener(ILogListener<LogEntry> listener)
	{
		if(parent != null)
		{
			parent.addListener(listener);
			return;
		}
		
		accessProtection.lock();
		try
		{
			listeners.add(listener);
			String[] logEntryStrings = readBuffer();
			if(logEntryStrings == null)
				return;
			
			List<LogEntry> logEntries = new LinkedList<LogEntry>();
			for(String entry : logEntryStrings)
			{
				LogEntry logEntry = getLogEntry(entry);
				if(logEntry != null)
					logEntries.add(logEntry);			
			}
			listener.log(logEntries);		    
		}
		finally
		{
			accessProtection.unlock();
		}
	}

	/**
	 * @param entry
	 * @return
	 */
	protected LogEntry getLogEntry(String entryString)
	{
		LogEntry entry = getNewLogEntry();
		entry.fillFromEntryTemplate(entryString, levelLabels);
		entry.setLogMessage(levelLabels);
		return entry;
	}

	/**
	 * @return
	 */
	protected String[] readBuffer()
	{
		if (localLogFileName == null)
			return new String[0];
		File file = new File(localLogFileName);
		if (!file.exists())
			return new String[0];
		if(!closeLocalLogWriter())
			return new String[0];
		
		String[] buffer = primReadBuffer();
		
		try
		{
			openLocalLogWriter();
		}
		catch (IOException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			return null;
		}
		
		return buffer;
	}

	protected String[] primReadBuffer()
	{	
		List<String> entries = new StringList();
		try
		{
			FileReader reader = new FileReader(localLogFileName);
			String entry = primReadNextLogEntry(reader);
			while(entry != null)
			{	
				if(entry.length() > 0 && !entry.equals("\n"))	// ignore crs //$NON-NLS-1$
				{
					if(StringUtility.contains(entry, NewSession))
						entries = new StringList();
					else	
						entries.add(entry);					
				}
				entry = primReadNextLogEntry(reader);
			}
			reader.close();
		}
		catch (FileNotFoundException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		catch (IOException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		return StringUtility.toArray(entries);
	}
	
	protected String primReadNextLogEntry(FileReader reader) throws IOException
	{
		StringBuffer str = new StringBuffer();
		int charNumber = reader.read();
		if(charNumber == -1)
			return null;
		char previous = (char) charNumber;
		while(charNumber != -1)
		{
			char ch = (char) charNumber;
			str.append(ch);
			if(ch == '>' && previous == '/')
				return str.toString();			
			previous = ch;
			charNumber = reader.read();
		}
		return str.toString();
	}

	protected boolean closeLocalLogWriter()
	{
		if (localLogWriter == null)
			return true;
		try
		{
			localLogWriter.close();
		}
		catch (IOException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			return false;
		}
		return true;
	}

	protected void openLocalLogWriter() throws IOException
	{
		localLogWriter = new FileWriter(localLogFileName, true);
	}

	public boolean removeListener(ILogListener listener)
	{
		if(parent != null)
			return parent.removeListener(listener);
		
		synchronized (this)
		{
			return listeners.remove(listener);
		}
	}
	
	public boolean removeSubscription(String eventName)
	{
		return subscribedEvents.remove(eventName);
	}

	public boolean changeLocalLogFileName(String newFilePathName) throws IOException
	{
		checkLocalLogMaxSize(newFilePathName);

		if (localLogWriter == null)
		{
			setLocalLogFileName(newFilePathName);
			return true;
		}

		if (newFilePathName.equals(localLogFileName))
			return true;

		accessProtection.lock();
		try
		{
		    localLogWriter.close();
			File logFile = new File(localLogFileName);
			File newLogFile = new File(newFilePathName);
			StringBuffer buffer = new StringBuffer();

			if (newLogFile.exists())
			{
				FileReader fr = new FileReader(logFile);
				if (logFile.length() > 0)
				{
					buffer.append("\n"); //$NON-NLS-1$
					buffer.append(NewSession + " started " + TimeUtility.now()); //$NON-NLS-1$
					buffer.append("\n\n"); //$NON-NLS-1$
					int ch = fr.read();
					while (ch != -1)
					{
						buffer.append((char) ch);
						ch = fr.read();
					}
				}
				fr.close();
				logFile.delete();
			}
			else
			{
				boolean success = logFile.renameTo(newLogFile);
				if (!success)
				{
					// re-open previous file writer
					localLogWriter = null;
					setLocalLogFileName(localLogFileName);
					return false;
				}
			}

			localLogWriter = new FileWriter(newLogFile, true);
			localLogFileName = newFilePathName;
			if (buffer.length() > 0)
				localLogWriter.write(buffer.toString());
			writeBuffer();
			return true;
		}
		finally
		{
		    accessProtection.unlock();
		}		
	}

	public void close()
	{
		if (localLogWriter != null)
		{
			accessProtection.lock();
			try
			{
				localLogWriter.close();
				localLogWriter = null;
			}
			catch (IOException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
			finally
			{
			    accessProtection.unlock();
			}			
		}
	}
	
	/**
	 * @return Return a new <code>LogEntry</code> object.
	 */
	public E getNewLogEntry()
	{
	    return (E) new LogEntry();
	}
}
