/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.ade.distributor;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.almendra.commons.events.IVetoableCallbackHandler;
import org.almendra.commons.events.IVetoableEventRegistry;
import org.almendra.commons.number.Long;
import org.almendra.commons.threads.IExecutable;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.collection.Association;
import org.almendra.commons.util.collection.ListUtility;
import org.almendra.commons.util.collection.MapUtility;
import org.almendra.commons.util.collection.StringList;
import org.almendra.commons.util.collection.StringSet;
import org.almendra.commons.util.swing.BillboardPanelController;
import org.almendra.janet.ade.commands.ExecuteWorkloadCommand;
import org.almendra.janet.ade.executor.AbstractAnchor;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.handlers.AskVetoShutdownEventParams;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.CapabilityPath;
import org.almendra.janet.cas.node.LocalAgentPath;
import org.almendra.janet.cas.node.LocalCapabilityPath;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.node.NodeName;
import org.almendra.janet.cas.node.SystemApplication;



public class DistributorAnchor extends AbstractAnchor
{
	public static final String NODE_LABEL = "AD Distributor"; //$NON-NLS-1$
	public static final String NODE_VIEW_PREFIX = "AD Distributor on"; //$NON-NLS-1$
	
	public static final String LABEL = "ADE_LOAD_DISTRIBUTOR_ANCHOR"; //$NON-NLS-1$
	public static final String CAPABILITY_NAME = "ADE_LOAD_DISTRIBUTOR"; //$NON-NLS-1$
	public static final String BALANCING_AGENT_NAME = "BALANCING_DISTRIBUTOR"; //$NON-NLS-1$
	public static final String SHARING_AGENT_NAME = "SHARING_DISTRIBUTOR"; //$NON-NLS-1$
	public static final String LOAD_ADMINISTRATOR_AGENT_NAME = "LOAD_ADMINISTRATOR"; //$NON-NLS-1$
	
	public static final String BILLBOARD_ATTRIBUTE_QUEUE_SIZE_CATEGORY = ".QueueSizeCategory"; //$NON-NLS-1$
	public static final String BILLBOARD_ATTRIBUTE_COMMANDS_WAITING_TIME = ".CommandsWaitingTime"; //$NON-NLS-1$
	
	public static final String LOG_EVENT_SHARING_COMMAND_DISTRIBUTED = "SHARING_COMMAND_DISTRIBUTED"; //$NON-NLS-1$
	public static final String LOG_EVENT_BALANCING_COMMAND_DISTRIBUTED = "BALANCING_COMMAND_DISTRIBUTED"; //$NON-NLS-1$
	public static final String LOG_EVENT_PANIC_CANNOT_FIND_ANY_AGENT = "PANIC_CANNOT_FIND_ANY_AGENT"; //$NON-NLS-1$
	
	protected Map<String, WorkstationLoadImage> workstationLoadImages = new HashMap<String, WorkstationLoadImage>();	//	<WorkstationNameAsString, WorkstationLoadImage>
	protected ReadWriteLock imagesLock = new ReentrantReadWriteLock(true);
	protected Logger logger = null;
	protected String lastSelectedSharingNodeName = null;
	protected long lastSelectedSharingNodeNameTime = -1;
	protected Descriptor descriptor = null;
	protected Node node = null;
	protected Map<String, List<String>> billboardMasks = new HashMap<String, List<String>>();	//	<NodeNameAsString, List<MaskAttributeAsString>>
	protected IAgentProxy loadAdministratorAgent = null;
	protected IAgentProxy balancerAgent = null;
	protected CapabilityQueueSizeStore waitingQueuesStore = new CapabilityQueueSizeStore();
	protected CapabilityQueueSizeStore executingQueuesStore = new CapabilityQueueSizeStore();
	protected Map<String, Long> agentsCountByCapability = new HashMap<String, Long>();	//	<CapabilityPathAsString, AgentCountAsLong>
	protected Map<String, WorkstationLoadImage> workstationLoadImageByNodeName = new HashMap<String, WorkstationLoadImage>();	//	<NodeNameAsString, WorkstationLoadImage>
	protected String lastSelectNodeName = null;
	
	/**
	 * Creates a new <code>DistributorAnchor</code> object.
	 */
	protected DistributorAnchor()
	{
		super();
	}
	
	/**
	 * Creates a new <code>DistributorAnchor</code> object.
	 */
	public DistributorAnchor(Node node, Logger parentLogger)
	{
		super();
		this.node = node;
		this.systemApplication = node.getSystemApplication();
		setLogger(parentLogger);
		registerForEvents();
	}

	/**
	 * 
	 */
	protected void registerForEvents()
	{
		IVetoableEventRegistry vetoableEventRegistry = node.getVetoableEventRegistry();
		IVetoableCallbackHandler handler = new AskVetoShutdownCallbackHandler(this);
		vetoableEventRegistry.register(AskVetoShutdownEventParams.EVENT_NAME, handler);
	}

	/**
	 * @param parentLogger
	 */
	protected void setLogger(Logger parentLogger)
	{
		logger = Logger.getChildLogger(parentLogger);
		logger.getDefaultValues().setAgentName(SHARING_AGENT_NAME);
		logger.getDefaultValues().setCapabilityName(CAPABILITY_NAME);
		logger.getDefaultValues().setApplicationName(SystemApplication.NAME);
		logger.registerForEvents(registeredLogEvents());
		logger.subscribeForEvents(registeredLogEvents());
	}

	/**
	 * @param i
	 * @return
	 */
	public String getLastSelectedSharingNode()
	{
		if(lastSelectedSharingNodeNameTime == -1)
			return null;
		if(lastSelectedSharingNodeName == null)
			return null;
		long timeNow = System.currentTimeMillis();
		if(timeNow - getSharingDelay() <= lastSelectedSharingNodeNameTime)
			return lastSelectedSharingNodeName;
		return null;
	}

	/**
	 * @param string
	 */
	public void setLastSelectedSharingNode(String nodeName)
	{
		lastSelectedSharingNodeName = nodeName;
		lastSelectedSharingNodeNameTime = System.currentTimeMillis();
	}
	
	/**
	 * @param desc
	 */
	public void setDescriptor(Descriptor descriptor)
	{
		this.descriptor = descriptor;
	}
	
	protected long getSharingDelay()
	{
		long defaultDelay = 2000;
		if(descriptor == null)
			return defaultDelay;
		long delay = descriptor.getSharingDelay();
		if(delay == -1)
			return defaultDelay;
		return delay;
	}
	
	public List<String> getAvailableNodes()
	{
		return getAvailableNodes(null);
	}
	
	public List<String> getAvailableNodes(String ignoreWorkstationNameOrNull)
	{
	    imagesLock.readLock().lock();
	    List<String> availableNodes = null;
	    try
	    {
	        availableNodes = new StringList();       	    
			for(WorkstationLoadImage workstation : getAvailableWorkstations())
			{
				if(ignoreWorkstationNameOrNull != null && !workstation.getWorkstationName().equals(ignoreWorkstationNameOrNull))
					availableNodes.addAll(workstation.getNodeNames());
			}
			return availableNodes;
	    }
	    finally
	    {
	        imagesLock.readLock().unlock();	        
	    }						
	}
	
	/**
	 * @return
	 */
	public List<WorkstationLoadImage> getAvailableWorkstations()
	{
	    imagesLock.readLock().lock();
	    List<WorkstationLoadImage> availableWorkstations = null;
		try
		{
			availableWorkstations = new ArrayList<WorkstationLoadImage>();			
			for(WorkstationLoadImage image : workstationLoadImages.values())
			{
				if(!image.isAboveThreshold())
					availableWorkstations.add(image);
			}
			return availableWorkstations;
		}
		finally
		{
			imagesLock.readLock().unlock();
		}
	}
	
	public void addWorkstationLoadImage(WorkstationLoadImage image)
	{
	    imagesLock.writeLock().lock();
	    try
		{	        
			AssertUtility.assertFalse(workstationLoadImages.containsKey(image.getWorkstationName()));
			workstationLoadImages.put(image.getWorkstationName(), image);  
		}
	    finally
	    {
	        imagesLock.writeLock().unlock();
	    }				
	}
	
	/**
	 * @param nodeName
	 * @param newQueueSizeCategory
	 */
	protected void displayQueueSizeCategory(String nodeName, int newQueueSizeCategory)
	{
		displayValue(nodeName, BILLBOARD_ATTRIBUTE_QUEUE_SIZE_CATEGORY, String.valueOf(newQueueSizeCategory));
	}
	
	/**
	 * @param nodeName
	 * @param newQueueSizeCategory
	 */
	protected void displayValue(String nodeName, String attribute, String value)
	{
		BillboardPanelController billboard = node.getBillboard();
		if(billboard == null)
			return;
		String fullAttributeName = nodeName + attribute;
		List<String> mask = billboardMasks.get(nodeName);
		if(mask == null)
		{
			mask = new StringList();
			if(billboardMasks.isEmpty())
				mask.add(BillboardPanelController.BLANK_LINE);
			mask.add(nodeName + BILLBOARD_ATTRIBUTE_QUEUE_SIZE_CATEGORY);
			AssertUtility.assertTrue(mask.contains(fullAttributeName));
			billboardMasks.put(nodeName, mask);
			billboard.setMask(nodeName, mask);
		}
		billboard.setValue(nodeName, fullAttributeName, value);
	}

	public int getQueueSizeCategory(String nodeName)
	{
	    imagesLock.readLock().lock();
		try
		{			
			NodeLoadImage nodeLoadImage = getNodeLoadImage(nodeName);
			return nodeLoadImage.getQueueSizeCategory();
		}
		finally
		{
			imagesLock.readLock().unlock();
		}
	}
		
	public WorkstationLoadImage getWorkstationLoadImage(String workstationName)
	{
	    imagesLock.writeLock().lock();
	    try
		{	        
	        WorkstationLoadImage workstationLoadImage = workstationLoadImages.get(workstationName);
	        if(workstationLoadImage == null)
	        {			
				workstationLoadImage = new WorkstationLoadImage(this, workstationName);
				addWorkstationLoadImage(workstationLoadImage);				
			}
	        return workstationLoadImage;
		}
		finally
		{
			imagesLock.writeLock().unlock();
		}						
	}
	
	public NodeLoadImage getNodeLoadImage(String nodeName)
	{
		WorkstationLoadImage workstationLoadImage = workstationLoadImageByNodeName.get(nodeName); 
		return workstationLoadImage.getNodeLoadImage(nodeName);
	}

	/**
	 * @param localAgentPath
	 * @return
	 */
	public Map<AgentPath, Long> getAgentsQueueSizes(LocalAgentPath localAgentPath)
	{
		return node.getClusterImage().getAgentsQueueSizes(localAgentPath);
	}

	/**
	 * @param string
	 * @return
	 */
	protected int getWorkstationQueueSizeCategoryProduct(String nodeName)
	{
	    imagesLock.readLock().lock();
		try
		{			
			WorkstationLoadImage workstationLoadImage = getWorkstationLoadImage(new NodeName(nodeName).getWorkstationName());
			return workstationLoadImage.getProduct();
		}
		finally
		{
			imagesLock.readLock().unlock();
		}
	}

	/**
	 * @param string
	 * @return
	 */
	protected int getNodeQueueSizeCategory(String nodeName)
	{
	    imagesLock.readLock().lock();
		try
		{			
			return getNodeLoadImage(nodeName).getQueueSizeCategory();
		}
		finally
		{
			imagesLock.readLock().unlock();
		}
	}

	/**
	 * @param string
	 * @param i
	 */
	public void setNonJanetCPULoad(String workstationName, int nonJanetCPULoad)
	{
	    imagesLock.writeLock().lock();
	    try
		{	        
	        WorkstationLoadImage image = getWorkstationLoadImage(workstationName);
	        image.setNonJanetCPULoad(nonJanetCPULoad);	        
		}
	    finally
	    {
	        imagesLock.writeLock().unlock();
	    }
	}
	
	/**
	 * @param string
	 * @param i
	 */
	public void setIsAboveThreshold(String workstationName, boolean isAboveThreshold, int nonJanetCPULoad)
	{
	    imagesLock.writeLock().lock();
		try
		{		    
		    WorkstationLoadImage image = getWorkstationLoadImage(workstationName);
		    image.setAboveThreshold(isAboveThreshold);
		    image.setNonJanetCPULoad(nonJanetCPULoad);
		}
		finally
		{
			imagesLock.writeLock().unlock();
		}
	}
	
	/**
	 * @param string
	 * @param i
	 */
	public int setNewNonJanetCPULoad(String workstationName, int nonJanetCPULoad)
	{
	    imagesLock.writeLock().lock();
		try
		{			
			WorkstationLoadImage image = getWorkstationLoadImage(workstationName);
			int previousNonJanetCPULoad = image.getNonJanetCPULoad(); 
			image.setNonJanetCPULoad(nonJanetCPULoad);
			return previousNonJanetCPULoad;
		}
		finally
		{
			imagesLock.writeLock().unlock();
		}
	}

	/**
	 * @param string
	 * @param currentQueueSizeCategory
	 * @param currentCommandsWaitingTime
	 */
	public void setNewNodeValues(String nodeName, int newQueueSizeCategory, long waitingTime, Map<String, Long> waitingQueuesByCapability, Map<String, Long> executingQueuesByCapability)
	{
	    imagesLock.writeLock().lock();
		try
		{			
			NodeLoadImage nodeLoadImage = getNodeLoadImage(nodeName);
			if(nodeLoadImage == null)
			{
				System.out.println("NodeLoadImage for " + nodeName + " not found!"); //$NON-NLS-1$ //$NON-NLS-2$
				return;
			}
			nodeLoadImage.setQueueSizeCategory(newQueueSizeCategory);
			nodeLoadImage.setWaitingTime(waitingTime);
			displayQueueSizeCategory(nodeName, newQueueSizeCategory);
			waitingQueuesStore.update(waitingQueuesByCapability, nodeName);
			executingQueuesStore.update(executingQueuesByCapability, nodeName);
		}
		finally
		{
			imagesLock.writeLock().unlock();
		}
	}
	
	/**
	 * @return Returns the loadAdministratorAgent.
	 */
	public IAgentProxy getBalancerAgent()
	{
		if(balancerAgent == null)
		{			
			try
			{
				AgentPath agentPath = new AgentPath(AgentPath.LOCAL_NODE_TAG, SystemApplication.NAME, CAPABILITY_NAME, BALANCING_AGENT_NAME);
				balancerAgent = node.getSystemApplication().getAgent(agentPath);
			}
			catch (Exception e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				node.getLogger().exception(e);
			}
		}
		return balancerAgent;
	}
	
	/**
	 * @return Returns the loadAdministratorAgent.
	 */
	public IAgentProxy getLoadAdministratorAgent()
	{
		if(loadAdministratorAgent == null)
		{			
			try
			{
				AgentPath agentPath = new AgentPath(AgentPath.LOCAL_NODE_TAG, SystemApplication.NAME, CAPABILITY_NAME, LOAD_ADMINISTRATOR_AGENT_NAME);
				loadAdministratorAgent = node.getSystemApplication().getAgent(agentPath);
			}
			catch (Exception e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				node.getLogger().exception(e);
			}
		}
		return loadAdministratorAgent;
	}
	
	/**
	 * @return Returns the node.
	 */
	public Node getNode()
	{
		return node;
	}
	/**
	 * @return Returns the logger.
	 */
	public Logger getLogger()
	{
		return logger;
	}
	
	public boolean writeProtectedDo(IExecutable executable)
	{
	    imagesLock.writeLock().lock();
		try
		{			
			executable.execute(this);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
			return false;
		}
		finally
		{
			imagesLock.writeLock().unlock();
		}
		return true;
	}

	/**
	 * @param workstationName
	 * @return
	 */
	public List<IAgentProxy> getAllSystemAgents(String workstationName)
	{
	    imagesLock.readLock().lock();
		try
		{			
			List<IAgentProxy> agentProxies = new ArrayList<IAgentProxy>();
			List<String> nodeNames = getWorkstationLoadImage(workstationName).getNodeNames();
			for(String nodeName : nodeNames)
			{
				agentProxies.add(node.getSystemApplication().getSystemAgent(nodeName));				
			}
			return agentProxies;
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().unlock();
		}
		
		return new ArrayList<IAgentProxy>();
	}
	/**
	 * @return Returns the executingQueuesStore.
	 */
	public CapabilityQueueSizeStore getExecutingQueuesStore()
	{
		return executingQueuesStore;
	}
	/**
	 * @return Returns the waitingQueuesStore.
	 */
	public CapabilityQueueSizeStore getWaitingQueuesStore()
	{
		return waitingQueuesStore;
	}

	/**
	 * @param nodeName
	 * @param capabilityPath
	 * @return
	 */
	public long getAgentsCount(String nodeName, String capabilityPath)
	{
		String key = capabilityPath + "." + nodeName; //$NON-NLS-1$
		Long countObject = agentsCountByCapability.get(key);
		if(countObject == null)
		{
			int count = node.getSystemApplication().getAgentCount(nodeName, capabilityPath);
			countObject = new Long(count);
			agentsCountByCapability.put(key, countObject);
		}
		return countObject.longValue();
	}

	/**
	 * 
	 */
	public void clearCaches()
	{
		agentsCountByCapability = new HashMap<String, Long>();
	}

	/**
	 * @param string
	 */
	public void registerWorkstation(String workstationName, List<String> nodeNames)
	{
	    imagesLock.writeLock().lock();
		try
		{			
			WorkstationLoadImage image = getWorkstationLoadImage(workstationName);
			for(String nodeName : nodeNames)
				image.addNodeLoadImage(nodeName);	
		}
		finally
		{
			imagesLock.writeLock().unlock();
		}								
	}

	/**
	 * @param string
	 */
	public void deregisterWorkstation(String workstationName, List<String> nodeNames)
	{
	    imagesLock.writeLock().lock();
		try
		{			
			WorkstationLoadImage workstationLoadImage = getWorkstationLoadImage(workstationName);
			if(workstationLoadImage != null)
				workstationLoadImage.removeNodeLoadImagesNamed(nodeNames);
			if(!workstationLoadImage.hasNodes())
				workstationLoadImages.remove(workstationName);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.writeLock().unlock();
		}		
	}

	/**
	 * @param string
	 * @param string2
	 * @return
	 */
	public Set<String> getAllNodesWithCapability(LocalCapabilityPath path, Set<String> nodeNamestoConsider)
	{
		return systemApplication.getAllNodeNamesWithCapability(path, nodeNamestoConsider);
	}

	/**
	 * @param nodeNames
	 * @return
	 */
	public List<NodeLoadImage> getLeastLoadedCapabilitiesAtLeastOne(Set<String> nodeNamesOfNodesWithCapability, LocalCapabilityPath path, LeastLoadedNodeDecisionInformation decisionInformation)
	{				
		imagesLock.readLock().lock();
		try
		{
		    decisionInformation.setNodeNamesWithCapability(nodeNamesOfNodesWithCapability);
			if(nodeNamesOfNodesWithCapability.isEmpty())
				return null;
			List<NodeLoadImage> nodeImages = getNodeImages(nodeNamesOfNodesWithCapability);
			if(nodeImages.isEmpty())
				return null;
			if(nodeImages.size() == 1)
			    return nodeImages;
			List<NodeLoadImage> availableNodeLoadImages = getAvailableNodeLoadAtLeastOne(nodeImages, decisionInformation);
			if(availableNodeLoadImages.size() == 1)
				return ListUtility.toListFromSingleObject(availableNodeLoadImages.get(0));
			return getLeastLoadedNodes(availableNodeLoadImages, path, decisionInformation);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().unlock();
		}		
		return null;		
	}

	/**
	 * @param availableNodeLoadImages
	 * @param path
	 * @return node name of least loaded node for capability
	 */
	protected List<NodeLoadImage> getLeastLoadedNodes(List<NodeLoadImage> availableNodeLoadImages, LocalCapabilityPath path, LeastLoadedNodeDecisionInformation decisionInformation)
	{
		if(availableNodeLoadImages.size() == 1)
			return ListUtility.toListFromSingleObject(availableNodeLoadImages.get(0));
		
		Map<NodeLoadImage, Integer> workstationQSCByNode = new HashMap<NodeLoadImage, Integer>();
		for(NodeLoadImage nodeLoadImage : availableNodeLoadImages)
		{
			int qsc = nodeLoadImage.getWorkstationLoadImage().getQSCProduct();
			decisionInformation.addWorkstationQSCByNodeName(nodeLoadImage.getNodeName(), new Integer(qsc));
			workstationQSCByNode.put(nodeLoadImage, new Integer(qsc));
		}
		
		List<NodeLoadImage> nodesWithSmallestWorkstationQSC = MapUtility.getKeysWithSmallestIntegerValue(workstationQSCByNode);
		decisionInformation.addNodeNamesWithSmallestWorkstationQSC(NodeLoadImage.getNodeNames(nodesWithSmallestWorkstationQSC));
		
		Map<NodeLoadImage, Integer> nodeQSCByNode = new HashMap<NodeLoadImage, Integer>();
		for(NodeLoadImage nodeLoadImage : nodesWithSmallestWorkstationQSC)
			nodeQSCByNode.put(nodeLoadImage, new Integer(nodeLoadImage.getQueueSizeCategory()));

		List<NodeLoadImage> nodesWithSmallestOverAllQSC = MapUtility.getKeysWithSmallestIntegerValue(nodeQSCByNode);
		decisionInformation.addNodeNamesWithSmallestOverAllQSC(NodeLoadImage.getNodeNames(nodesWithSmallestOverAllQSC));
		
		if(nodesWithSmallestOverAllQSC.size() == 1)
		{
			NodeLoadImage nodeLoadImage = nodesWithSmallestOverAllQSC.get(0);
			decisionInformation.setNodeNameWithSmallestOverAllQSC(nodeLoadImage.getNodeName());
			return ListUtility.toListFromSingleObject(nodeLoadImage);
		}
		
		Map<NodeLoadImage, Integer> cqsByNode = new HashMap<NodeLoadImage, Integer>();
		for(NodeLoadImage nodeLoadImage : nodesWithSmallestOverAllQSC)
		{
			int cqs = nodeLoadImage.getCapabilityQueueSize(path);
			decisionInformation.addCQSByNode(nodeLoadImage.getNodeName(), new Integer(cqs));
			cqsByNode.put(nodeLoadImage, new Integer(cqs));
		}
		
		List<NodeLoadImage> leastLoadedNodes = MapUtility.getKeysWithSmallestIntegerValue(cqsByNode);
		List<String> leastLoadedNodeNames = NodeLoadImage.getNodeNames(leastLoadedNodes);
		decisionInformation.setLeastLoadedNodes(leastLoadedNodeNames);
		return leastLoadedNodes;
	}

	/**
	 * @param nodeImages
	 * @return
	 */
	protected List<NodeLoadImage> getAvailableNodeLoadAtLeastOne(List<NodeLoadImage> nodeImages, LeastLoadedNodeDecisionInformation decisionInformation)
	{
		List<NodeLoadImage> availableNodeLoadImages = new ArrayList<NodeLoadImage>();
		List<String> availableNodeNames = new StringList();
		List<NodeLoadImage> notAvailableNodeLoadImages = new ArrayList<NodeLoadImage>();
		for(NodeLoadImage nodeLoadImage : nodeImages)
		{
			if(nodeLoadImage.isAboveThreshold())
				notAvailableNodeLoadImages.add(nodeLoadImage);
			else
			{
				availableNodeLoadImages.add(nodeLoadImage);
				availableNodeNames.add(nodeLoadImage.getNodeName());
			}
		}
		decisionInformation.setAvailableNodeNames(availableNodeNames);
		if(availableNodeLoadImages.size() >= 1)
			return availableNodeLoadImages;
		
		int minNonJanetCPULoad = Integer.MAX_VALUE;
		NodeLoadImage leastLoadedNotAvailableNodeLoadImage = null;
		
		for(NodeLoadImage nodeLoadImage : notAvailableNodeLoadImages)
		{
			if(nodeLoadImage.getNonJanetCPULoad() < minNonJanetCPULoad)
			{
				leastLoadedNotAvailableNodeLoadImage = nodeLoadImage;
				minNonJanetCPULoad = nodeLoadImage.getNonJanetCPULoad();
			}
		}
		decisionInformation.setLeastLoadedNotAvailableNodeName(leastLoadedNotAvailableNodeLoadImage.getNodeName());
		return ListUtility.toListFromSingleObject(leastLoadedNotAvailableNodeLoadImage);
	}

	/**
	 * @param nodeNames
	 * @return
	 */
	protected List<NodeLoadImage> getNodeImages(Set<String> nodeNames)
	{
		List<NodeLoadImage> nodeLoadImages = new ArrayList<NodeLoadImage>();
		for(String nodeName : nodeNames)
		{
			WorkstationLoadImage workstationLoadImage = workstationLoadImageByNodeName.get(nodeName);
			if(workstationLoadImage != null)
			{
				NodeLoadImage nodeLoadImage = workstationLoadImage.getNodeLoadImage(nodeName);
				nodeLoadImages.add(nodeLoadImage);
			}
			else
			{
				System.out.println(getClass().getName() + ".getNodeImages: no workstationLoadImage for node " + nodeName); //$NON-NLS-1$
			}
		}		
		return nodeLoadImages;
	}

	/**
	 * @param nodeNames
	 * @return
	 */
	protected List<WorkstationLoadImage> getWorkstationImages(List<String> nodeNames)
	{
	    imagesLock.writeLock().lock();
	    try
	    {	       
	        List<WorkstationLoadImage> nodeImages = new ArrayList<WorkstationLoadImage>();
	        for(String nodeName : nodeNames)
	            nodeImages.add(workstationLoadImageByNodeName.get(nodeName));
	        return nodeImages;
	    }
	    finally
	    {
	        imagesLock.writeLock().unlock();
	    }
	}

	/**
	 * @param leastLoadedNodeForCapability
	 * @param originAgentPath
	 * @return
	 */
	public AgentPath getAgentPathToNode(String nodeName, AgentPath originAgentPath)
	{
		CapabilityPath path = new CapabilityPath(nodeName, originAgentPath.getApplicationName(), originAgentPath.getCapabilityName()); 
		List<AgentPath> agentPaths = systemApplication.getAllAgentPaths(path);
		return agentPaths.get(0);
	}

	/**
	 * @param command
	 * @param path
	 * @param selectedAgentPath
	 */
	public void logCommandDistributed(ExecuteWorkloadCommand command, AgentPath originAgentPath, AgentPath selectedAgentPath)
	{
		LogEntry entry = logger.getNewLogEntry();
		String eventName = command.getLogEventName();
		String typeName = command.getLogTypeName();
		entry.setEvent(eventName);
		StringBuffer buffer = new StringBuffer();
		buffer.append("distributed " + typeName + command.toLogString() + " from " + originAgentPath.getQualifiedPathName()); //$NON-NLS-1$ //$NON-NLS-2$
		buffer.append(" to " + selectedAgentPath.getQualifiedPathName()); //$NON-NLS-1$
		entry.setArgument(buffer.toString());
		logger.info(entry);
	}

	/**
	 * @param nodeName
	 * @param path
	 * @return
	 */
	public int getCapabilityQueueSize(String nodeName, String qualifiedCapabilityPath)
	{
		return waitingQueuesStore.getQueueSize(nodeName, qualifiedCapabilityPath) + executingQueuesStore.getQueueSize(nodeName, qualifiedCapabilityPath);
	}
	
	/**
	 * @param nodeName
	 * @param path
	 * @return
	 */
	public int getCapabilityQueueSize(String nodeName, LocalCapabilityPath path)
	{
		return waitingQueuesStore.getQueueSize(nodeName, path.getQualifiedPathName()) + executingQueuesStore.getQueueSize(nodeName, path.getQualifiedPathName());
	}

	/**
	 * @param string
	 * @param nodeName
	 * @return
	 */
	public List<LocalCapabilityPath> getAlternativeCapabilitiesOnOtherNodes(String workstationName, String nodeName)
	{
		List<String> nodeNames = getWorkstationLoadImage(workstationName).getNodeNames();
		List<LocalCapabilityPath> localCapabilityPaths = systemApplication.getAllApplicationCapabilitesOn(nodeName);
		return systemApplication.existingCapabilitiesOnOtherNodes(localCapabilityPaths, nodeNames);
	}

	/**
	 * @return
	 */
	public List<String> getAvailableNodeNamesWithQSC0()
	{
	    imagesLock.readLock().lock();
		try
		{			
			List<String> nodeNamesWithQSC0 = new StringList();
			for(WorkstationLoadImage workstationLoadImage : workstationLoadImages.values())
				nodeNamesWithQSC0.addAll(workstationLoadImage.getAvailableNodeNamesWithQSC0());
			return nodeNamesWithQSC0;
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().unlock();
		}		
		return new StringList();
	}

	/**
	 * @param nodeNameAboveQSC1
	 * @param nodeNameWithQSC0
	 * @return
	 */
	public List<LocalCapabilityPath> getCapabilityIntersection(String nodeName1, String nodeName2)
	{
		List<LocalCapabilityPath> localCapabilityPaths1 = systemApplication.getAllApplicationCapabilitesOn(nodeName1);
		List<LocalCapabilityPath> localCapabilityPaths2 = systemApplication.getAllApplicationCapabilitesOn(nodeName2);
		return ListUtility.getIntersection(localCapabilityPaths1, localCapabilityPaths2);
	}
	
	/**
	 * @param nodeNameToMigrateTo name of node to migrate command to
	 * @param capabilities
	 * @return name of node to evict command from
	 */
	public String getNodeNameToEvictCommandFrom(String nodeNameToMigrateTo, Set<String> nodeNamesAboveQSC1, Set<String> capabilities, NodeDroppedToQSC1DecisionInformation decisionInformation)
	{
		List<LocalCapabilityPath> destinationCapabilities = systemApplication.getAllApplicationCapabilitesOn(nodeNameToMigrateTo);
		decisionInformation.setDestinationCapabilities(LocalCapabilityPath.getQualifiedPathNames(destinationCapabilities));
		Map<String, Set<String>> capabilitiesByCandidateNodeName = new HashMap<String, Set<String>>(); 
		Set<String> allCandidateNodes = new StringSet();
		for(LocalCapabilityPath path : destinationCapabilities)
		{
			Set<String> nodeNames = getAllNodesWithCapability(path, new StringSet(nodeNamesAboveQSC1));
			if(!nodeNames.isEmpty())
			{
				MapUtility.putAllStrings(capabilitiesByCandidateNodeName, path.getQualifiedPathName(), nodeNames);
				allCandidateNodes.addAll(nodeNames);
			}
		}
		if(capabilitiesByCandidateNodeName.isEmpty())
			return null;
		decisionInformation.setCandidateNodeNamesWithRequiredCapabilities(allCandidateNodes);
		String heaviestLoadedNodeName = getHeaviestLoadedNodeName(allCandidateNodes, capabilitiesByCandidateNodeName, capabilities);
		decisionInformation.setHeaviestLoadedNodeName(heaviestLoadedNodeName);
		capabilities.addAll(capabilitiesByCandidateNodeName.get(heaviestLoadedNodeName));
		decisionInformation.setCapabilitiesOfHeaviestLoadedNode(capabilities);
		return heaviestLoadedNodeName;
	}

	/**
	 * @param allCandidateNodes
	 * @param capabilitiesByCandidateNodeName
	 * @return
	 */
	protected String getHeaviestLoadedNodeName(Set<String> allCandidateNodes, Map<String, Set<String>> capabilitiesByCandidateNodeName, Set<String> heaviestLoadedCapabilityPaths)
	{
		Map<String, Integer> workstationQSCByNodeName = new HashMap<String, Integer>();
		for(String nodeName : allCandidateNodes)
		{
			int workstationQSC = getNodeLoadImage(nodeName).getWorkstationLoadImage().getQSCProduct();
			workstationQSCByNodeName.put(nodeName, new Integer(workstationQSC));
		}
		List<String> smallestWorkstationQSCNodeNames = MapUtility.getKeysWithSmallestIntegerValue(workstationQSCByNodeName);
		Map<String, Integer> nodeQSCByNodeName = new HashMap<String, Integer>();
		for(String nodeName : smallestWorkstationQSCNodeNames)
		{
			int nodeQSC = getNodeLoadImage(nodeName).getQueueSizeCategory();
			nodeQSCByNodeName.put(nodeName, new Integer(nodeQSC));
		}
		List<String> smallestNodesQSCNodeNames = MapUtility.getKeysWithSmallestIntegerValue(nodeQSCByNodeName);
		Map<Association<String, String>, Integer> smallestCQSByNodeCapabilityPathAssoc = new HashMap<Association<String, String>, Integer>();
		for(String nodeName : smallestNodesQSCNodeNames)
		{
			NodeLoadImage nodeLoadImage = getNodeLoadImage(nodeName);
			Set<String> capabilities = capabilitiesByCandidateNodeName.get(nodeName);
			for(String capabilityPath : capabilities)
			{
				int cqs = nodeLoadImage.getCapabilityQueueSize(capabilityPath);
				smallestCQSByNodeCapabilityPathAssoc.put(new Association<String, String>(nodeName, capabilityPath), new Integer(cqs));
			}
		}
		Association<String, String> smallestNodeCapabilityPathAssocs = MapUtility.getKeyWithGreatestIntegerValue(smallestCQSByNodeCapabilityPathAssoc);
		if(smallestNodeCapabilityPathAssocs == null)
			return null;
		String heaviestLoadedNode = smallestNodeCapabilityPathAssocs.getKey();
		String heaviestLoadedCapabilityPath = smallestNodeCapabilityPathAssocs.getValue();
		heaviestLoadedCapabilityPaths.add(heaviestLoadedCapabilityPath);
		return heaviestLoadedNode;
	}

	/**
	 * @param nodeNameDroppedToQSC1
	 * @param capabilityPath
	 * @return
	 */
	public AgentPath getAnyAgent(String nodeName, String capabilityPath)
	{
		LocalCapabilityPath path = new LocalCapabilityPath(capabilityPath);
		List<AgentPath> agentPaths = systemApplication.getAllAgentPaths(new CapabilityPath(nodeName, path.getApplicationName(), path.getCapabilityName()));
		if(agentPaths == null || agentPaths.isEmpty())
			return null;
		return agentPaths.get(0);
	}

	/**
	 * @param nodeNameAboveQSC1
	 * @param capabilitiesIntersection
	 * @return
	 */
	public String getHeaviestLoadedCapabilityPath(String nodeName, List<LocalCapabilityPath> capabilities)
	{
		// TODO - refactor getHeaviestLoadedNode to use this method
		NodeLoadImage nodeLoadImage = getNodeLoadImage(nodeName);
		if(nodeLoadImage == null)
			return null;
		Map<String, Integer> cqsByCapabilityPath = new HashMap<String, Integer>();
		for(LocalCapabilityPath capabilityPath : capabilities)
		{
			int cqs = nodeLoadImage.getCapabilityQueueSize(capabilityPath.getQualifiedPathName());
			cqsByCapabilityPath.put(capabilityPath.getQualifiedPathName(), new Integer(cqs));
		}
		return MapUtility.getKeyWithGreatestIntegerValue(cqsByCapabilityPath);
	}
	
	public Set<String> getAllConnectedNodeNames()
	{
		return workstationLoadImageByNodeName.keySet();
	}

	public List<String> getAvailableNodeNamesAboveQSC(int qsc)
	{
	    imagesLock.readLock().lock();
	    List<String> nodeNames = null;
		try
		{
			nodeNames = new StringList();			
			for(WorkstationLoadImage workstationLoadImage : workstationLoadImages.values())
			{
				if(workstationLoadImage.isBelowThreshold())
				{
					for(NodeLoadImage nodeLoadImage : workstationLoadImage.getNodeLoadImages())
					{
						if(nodeLoadImage.getQueueSizeCategory() > qsc)
							nodeNames.add(nodeLoadImage.getNodeName());
					}
				}
			}
			return nodeNames;
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		finally
		{
			imagesLock.readLock().unlock();
		}
		return new StringList();
	}
	
	/**
	 * @return
	 */
	public boolean existsNodeAboveQSC0()
	{
		return !getAvailableNodeNamesAboveQSC(0).isEmpty();
	}
	
	/**
	 * @return
	 */
	public boolean isDisableEviction()
	{
		return descriptor.isDisableEviction();
	}
	/**
	 * @return Returns the lastSelectedNodeName.
	 */
	public String getLastSelectNodeName()
	{
		return lastSelectNodeName;
	}
	/**
	 * @param lastSelectedNodeName The lastSelectedNodeName to set.
	 */
	public void setLastSelectNodeName(String lastSelectNodeName)
	{
		this.lastSelectNodeName = lastSelectNodeName;
	}

	/**
	 * @param nodeNameHeaviestLoaded
	 * @param capabilities
	 * @return
	 */
	public String getHeaviestLoadedCapability(String nodeNameHeaviestLoaded, Set<String> capabilityPathNames, NodeDroppedToQSC1DecisionInformation decisionInformation)
	{
		Map<String, Integer> cqsByCapabilityName = new HashMap<String, Integer>();
		for(String capabilityPathName : capabilityPathNames)
		{
			int cqs = getCapabilityQueueSize(nodeNameHeaviestLoaded, capabilityPathName);
			cqsByCapabilityName.put(capabilityPathName, new Integer(cqs));
		}
		decisionInformation.setCQSByCapabilityName(cqsByCapabilityName);
		// TODO - improve not to pick an element at random if cqsByCapabilityName.size() > 1 
		String heaviestLoadedCapabilityName = MapUtility.getKeyWithGreatestIntegerValue(cqsByCapabilityName);
		decisionInformation.setHeaviestLoadedCapabilityName(heaviestLoadedCapabilityName);
		return heaviestLoadedCapabilityName;
	}

	/**
	 * @return
	 */
	public boolean isDisableFullEviction()
	{
		return descriptor.isDisableFullEviction();
	}
	/**
	 * @return
	 */
	public boolean isDisableNodeDroppedToQSC1()
	{
		return descriptor.isDisableNodeDroppedToQSC1();
	}
	/**
	 * @return
	 */
	public boolean isDisableNodesWithQSC0Exist()
	{
		return descriptor.isDisableNodesWithQSC0Exist();
	}
	/**
	 * @return
	 */
	public boolean isDisablePartialEviction()
	{
		return descriptor.isDisablePartialEviction();
	}
	/**
	 * @return Returns the descriptor.
	 */
	public Descriptor getDescriptor()
	{
		return descriptor;
	}

	/**
	 * @return
	 */
	public boolean hasWorkstations()
	{
	    imagesLock.readLock().lock();
		try
		{			
			return !workstationLoadImages.isEmpty();
		}
		finally
		{
			imagesLock.readLock().unlock();
		}
	}

	/**
	 * @return
	 */
	public Set<String> getWorkstationNames()
	{
	    imagesLock.readLock().lock();
		try
		{			
			return workstationLoadImages.keySet();
		}
		finally
		{
			imagesLock.readLock().unlock();
		}
	}

    
    /**
     * 
     */
    public void lockReadLock()
    {
        imagesLock.readLock().lock();
    }

    
    /**
     * 
     */
    public void unlockReadLock()
    {
        imagesLock.readLock().unlock();
    }
}
