/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.ade.executor.interpreters;

import java.io.File;
import java.io.IOException;
import java.util.List;

import org.almendra.commons.events.IEventHandler;
import org.almendra.commons.events.ISimpleEventRegistry;
import org.almendra.commons.util.collection.StringList;
import org.almendra.commons.util.xml.XMLParseException;
import org.almendra.janet.ade.commands.StartCommand;
import org.almendra.janet.ade.executor.Descriptor;
import org.almendra.janet.ade.executor.ExecutorAnchor;
import org.almendra.janet.ade.executor.handlers.CommandArrivedHandler;
import org.almendra.janet.ade.executor.handlers.NodeAboutToShutdownHandler;
import org.almendra.janet.ade.executor.handlers.NodeDeregisterHandler;
import org.almendra.janet.ade.executor.handlers.NodeRegisteredHandler;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.events.EventHandlerEnvelope;
import org.almendra.janet.cas.events.HandlerList;
import org.almendra.janet.cas.events.ILocalEventRegistry;
import org.almendra.janet.cas.handlers.CommandEnvelopeParams;
import org.almendra.janet.cas.handlers.NodeAboutToShutdownParams;
import org.almendra.janet.cas.handlers.NodeDeregisteredParams;
import org.almendra.janet.cas.handlers.NodeRegisteredParams;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.IExecutorAnchor;
import org.almendra.janet.cas.node.SystemApplication;
import org.almendra.janet.cas.scheduling.CommandAccessor;
import org.almendra.janet.cas.scheduling.IInterpreter;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.input.SAXBuilder;



public class StartInterpreter extends ExecutorInterpreter implements IInterpreter
{
	/**
	 * Creates a new <code>StartInterpreter</code> object.
	 */
	public StartInterpreter()
	{
		super();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#execute(org.almendra.commons.janet.cas.scheduling.CommandAccessor)
	 */
	public void execute(CommandAccessor cmdAccessor)
	{
		if(cmdAccessor.getCommand() instanceof StartCommand)
		{
			SystemApplication app = cmdAccessor.getNodeIfPrivilege().getSystemApplication();
			ExecutorAnchor anchor = new ExecutorAnchor(getSystemAgent(cmdAccessor), app, cmdAccessor.getLogger()); 
			
			if(!readQueueSizeCategories(cmdAccessor, anchor))
				return;
			
			cmdAccessor.getNodeObjectSpace().put(IExecutorAnchor.LABEL, anchor);
						
			registerForEvents(cmdAccessor, anchor);
		}
	}

	/**
	 * @param cmdAccessor
	 * @param anchor
	 */
	protected void registerForEvents(CommandAccessor cmdAccessor, ExecutorAnchor anchor)
	{
		ISimpleEventRegistry eventRegistry = cmdAccessor.getNodeIfPrivilege().getSimpleEventRegistry();
		IEventHandler handler = new CommandArrivedHandler(anchor);
		anchor.getHandlers().add(handler);
		eventRegistry.register(CommandEnvelopeParams.EVENT_NAME_COMMAND_ARRIVED, handler);	
		
		AgentPath systemAgentPath = cmdAccessor.getSystemApplicationIfPrivilege().getSystemAgentPath();
		
		ILocalEventRegistry systemEventRegistry = cmdAccessor.getNodeEventRegistry();
		handler = new NodeAboutToShutdownHandler(anchor);
		EventHandlerEnvelope env = new EventHandlerEnvelope(handler, systemAgentPath);
		HandlerList handlerList = new HandlerList(env);
		systemEventRegistry.register(NodeAboutToShutdownParams.EVENT_NAME, handlerList, cmdAccessor.getApplication());
		
		handler = new NodeDeregisterHandler(anchor);
		env = new EventHandlerEnvelope(handler, systemAgentPath);
		handlerList = new HandlerList(env);
		cmdAccessor.getNodeEventRegistry().register(NodeDeregisteredParams.EVENT_NAME, handlerList, cmdAccessor.getApplication());
		
		handler = new NodeRegisteredHandler(anchor);
		env = new EventHandlerEnvelope(handler, systemAgentPath);
		handlerList = new HandlerList(env);
		cmdAccessor.getNodeEventRegistry().register(NodeRegisteredParams.EVENT_NAME, handlerList, cmdAccessor.getApplication());
	}

	/**
	 * @param capability
	 * @return
	 */
	protected boolean readQueueSizeCategories(CommandAccessor cmdAccessor, ExecutorAnchor anchor) 
	{
		String descriptorFilePathName = getDescriptorFilePathName(cmdAccessor);
		if(descriptorFilePathName == null)
			return false;
		
		File file = getDescriptorFile(cmdAccessor, descriptorFilePathName);
		if(file == null)
			return false;
	
		SAXBuilder builder = new SAXBuilder();
		Document doc = null;
		
		try
		{
			doc = builder.build(file);
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.getLogger().exception(e);
			return false;
		}
		
		Element root = doc.getRootElement();
		Descriptor desc = null;
		
		try
		{
			desc = Descriptor.create(root);
			anchor.setDescriptor(desc);
		}
		catch (XMLParseException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.getLogger().exception(e);
			return false;
		}

		return true;
	}

	protected String getDescriptorFilePathName(CommandAccessor cmdAccessor)
	{
		String descriptorFilePathName = cmdAccessor.getCapability().getDescriptorFilePathName();
		if(descriptorFilePathName == null)
		{
			StringBuffer buffer = new StringBuffer();
			buffer.append("\n" + getClass().getName() + ".readQueueSizeCategories(...): "); //$NON-NLS-1$ //$NON-NLS-2$
			buffer.append("descriptor null!"); //$NON-NLS-1$
			buffer.append("\n" + cmdAccessor.getCapabilityName() + " cannot be started!\n"); //$NON-NLS-1$ //$NON-NLS-2$
			cmdAccessor.getLogger().fatalText(buffer.toString());
			return null;
		}
		return descriptorFilePathName;
	}
	
	/**
	 * @param cmdAccessor
	 * @param descriptorFilePathName
	 * @return
	 * @throws IOException
	 */
	protected File getDescriptorFile(CommandAccessor cmdAccessor, String descriptorFilePathName)
	{
		File file = new File(descriptorFilePathName);		
		if (!file.exists())
		{			
			String path = null;
			try
			{
				path = file.getCanonicalPath();
			}
			catch (IOException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				cmdAccessor.getLogger().exception(e);
				return null;
			}
			if(path != null)
				cmdAccessor.getLogger().fatalText("ADE descriptor " + descriptorFilePathName + " not found! Startup directory: " + path); //$NON-NLS-1$ //$NON-NLS-2$
		}
		return file;
	}

	/**
	 * @param cmdAccessor
	 * @return
	 * @throws NoSuchAgentException
	 * @throws IllegalAgentAccessException
	 * @throws InsufficientPrivilegeException
	 */
	protected IAgentProxy getSystemAgent(CommandAccessor cmdAccessor)
	{
		try
		{
			return cmdAccessor.getSystemAgent();
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			cmdAccessor.getLogger().exception(e);
		}
		return null;
	}
	
	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.scheduling.IInterpreter#commandNames()
	 */
	public List<String> commandNames()
	{
		return new StringList(StartCommand.QualifiedName);
	}
}
