/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.janet.cas.agents;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.almendra.commons.util.collection.CollectionUtility;
import org.almendra.commons.util.collection.ListUtility;
import org.almendra.commons.util.collection.StringList;
import org.almendra.commons.util.exception.InsufficientPrivilegeException;
import org.almendra.commons.util.globalId.IUniqueGlobalIdObject;
import org.almendra.commons.util.globalId.UniqueId;
import org.almendra.commons.util.globalId.UniqueIdGenerator;
import org.almendra.commons.util.logging.ILoggable;
import org.almendra.janet.cas.node.AbstractApplication;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.Capability;
import org.almendra.janet.cas.node.IllegalAgentAccessException;
import org.almendra.janet.cas.node.LocalAgentPath;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.NoSuchAgentException;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.node.SystemApplication;
import org.almendra.janet.cas.schedulers.CommandScheduler;
import org.almendra.janet.cas.scheduling.ICommandEnvelope;
import org.almendra.janet.cas.scheduling.IInterpreter;
import org.almendra.janet.cas.spaces.IObjectSpace;


/**
 * @author Oliver
 */
public abstract class AbstractAgent implements IAgent, ILoggable, IUniqueGlobalIdObject
{
	public static final String LOG_EVENT_STARTED_AGENT = "STARTED_AGENT"; //$NON-NLS-1$
	public static final String LOG_EVENT_STOPPED_AGENT = "STOPPED_AGENT"; //$NON-NLS-1$

	protected Node node = null;
	protected String name = null;
	protected Logger logger = null;
	protected Capability capability = null;
	protected UniqueId id = UniqueIdGenerator.next();
	
	protected AbstractAgent()
	{
		super();
		init();
	}

	protected AbstractAgent(Node node, Logger logger, String name)
	{
		this.node = node;
		this.name = name;
		this.logger = Logger.getChildLogger(logger);
		init();
	}
	
	protected String checkCapabilities(Set<Capability> capabilities)
	{
		if (capabilities == null)
			return "capabilities must not be null!"; //$NON-NLS-1$
		if (capabilities.isEmpty())
			return "capabilities must not empty!"; //$NON-NLS-1$

		List<List<IInterpreter>> interpreters = new ArrayList<List<IInterpreter>>();
		
		for(Capability cap : capabilities)
		{
			if(!cap.hasInterpreters())
				return "capability " + cap.getName() + " has no interpreters!"; //$NON-NLS-1$ //$NON-NLS-2$
			interpreters.add(cap.getInterpreters());
		}		
		
		StringBuffer buffer = new StringBuffer();
		Iterator<IInterpreter> it = CollectionUtility.duplicates(ListUtility.toList(interpreters)).iterator();
		while (it.hasNext())
		{
			IInterpreter interpreter = it.next();
			buffer.append(interpreter.getClass().getName());
			if (it.hasNext())
				buffer.append(", "); //$NON-NLS-1$
		}
		if (buffer.length() > 0)
		{
			return "capabilities must not share identical interpreters: " + buffer.toString(); //$NON-NLS-1$
		}
		return buffer.toString();
	}

	protected void init()
	{
		logger.getDefaultValues().setAgentName(getName());
		logger.registerForEvents(registeredLogEvents());
	}

	public void accept(ICommandEnvelope commandEnvelope)
	{
		setCommandEnvelope(commandEnvelope, this, getScheduler().getLogger(), node);
		getScheduler().accept(commandEnvelope);
	}

	public static void setCommandEnvelope(ICommandEnvelope commandEnvelope, IAgent agent, Logger logger, Node node)
	{
		commandEnvelope.setCurrentAgent(agent);
		commandEnvelope.setLogger(logger);
		commandEnvelope.setNode(node);
	}

	protected abstract CommandScheduler getScheduler();

	/* (non-Javadoc)
	 * @see org.almendra.commons.cas.node.INamedObject#getName()
	 */
	public String getName()
	{
		return name;
	}

	/**
	 * @param string
	 */
	public void setName(String name) throws IllegalArgumentException, NullPointerException
	{
		if (name == null)
			throw new NullPointerException("name is null"); //$NON-NLS-1$
		if (name.length() == 0)
			throw new IllegalArgumentException("no name specified"); //$NON-NLS-1$
		this.name = name;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.logging.ILoggable#registeredLogEvents()
	 */
	public List<String> registeredLogEvents()
	{
		List<String> events = new StringList();
		events.add(LOG_EVENT_STARTED_AGENT);
		return events;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.IUniqueGlobalIdObject#getUniqueId()
	 */
	public UniqueId getUniqueId()
	{
		return id;
	}

	/**
	 * @param logger
	 */
	public void setLogger(Logger logger)
	{
		this.logger = logger;
	}

	/**
	 * @param node
	 */
	protected void setNode(Node node)
	{
		this.node = node;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.globalId.IUniqueGlobalIdObject#getGlobalId()
	 */
	public UniqueId getGlobalId()
	{
		return id;
	}
	
	protected abstract void schedulerStopped();

	protected void logStarted()
	{
		LogEntry entry = LogEntry.getEntryTemplate(logger);
		entry.setEvent(LOG_EVENT_STARTED_AGENT);
		entry.setArgument("agent name/id={" + getName() + "/" + getGlobalId().getValue() + "}"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		logger.info(entry);
	}
	
	protected void logStopped()
	{
		LogEntry entry = LogEntry.getEntryTemplate(logger);
		entry.setEvent(LOG_EVENT_STOPPED_AGENT);
		entry.setArgument("agent name/id={" + getName() + "/" + getGlobalId()); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		logger.info(entry);
	}

	/**
	 * @return
	 */
	public Node getNode() throws InsufficientPrivilegeException
	{
		if(isSystemApplication())
			return node;
		throw new InsufficientPrivilegeException("privilege only granted to system agent"); //$NON-NLS-1$
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.cas.node.IAgent#getApplication()
	 */
	public AbstractApplication getApplication()
	{
		return capability.getApplication();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.cas.node.ISimpleActiveObject#isActive()
	 */
	public boolean isActive()
	{
		return getScheduler().isActive();
	}

	public IObjectSpace getObjectSpace()
	{
		return getApplication().getObjectSpace();
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.cas.node.IAgent#getQualifiedName()
	 */
	public String getQualifiedName()
	{
		return capability.getQualifiedName() + "." + getName(); //$NON-NLS-1$
	}

	/**
	 * @return
	 */
	public Capability getCapability()
	{
		return capability;
	}
	
	public abstract boolean isSystemAgent();
	public abstract boolean isApplicationAgent();

	/**
	 * 
	 */
	public SystemApplication getSystemApplication() 
	throws InsufficientPrivilegeException
	{
		if(isSystemApplication())
			return (SystemApplication) getApplication();
		throw new InsufficientPrivilegeException("privilege only granted to system agent"); //$NON-NLS-1$;
	}
	
	public IAgentProxy getAgentProxy() 
	{
		// Retrieving the agent's own proxy must always work.
		// If an exception occurs, there is some logical error.
		// So don't pass on any excpetion that occurs
		try
		{
			return getApplication().getAgent(getPath());
		}
		catch (NoSuchAgentException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		catch (IllegalAgentAccessException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			logger.exception(e);
		}
		return null;
	}
	
	public AgentPath getPath()
	{
		return new AgentPath(
			getApplication().getNode().getName(), 
			getApplication().getName(), 
			getCapability().getName(), 
			getName());
	}
	
	public LocalAgentPath getLocalPath()
	{
		return new LocalAgentPath(
			getApplication().getName(), 
			getCapability().getName(), 
			getName());
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.logger.node.IAgent#stop()
	 */
	public void stop()
	{
		// TODO - stop agent with brute force
		// kludge for now to make deregistration of applications work
		// need to make sure that schedulerAnchor has no more commands and that application
		// deregistered from all nodes in the cluster
		getScheduler().stop();
		schedulerStopped();				
	}

	protected void prepareFinalize()
	{
		node = null;
		logger = null;
		capability = null;
	}

	/**
	 * @return
	 */
	public IAgentProxy getAllForeignAgents() 
	throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{		
		return getApplication().getAllForeignAgents(getLocalPath());
	}

	/**
	 * @return
	 */
	public AgentPath getAgentPath()
	{
		return new AgentPath(node.getName(), getApplication().getName(), getCapability().getName(), getName());
	}

	/**
	 * @param string
	 * @return
	 */
	public IAgentProxy getAllForeignAgents(String capabilityName) 
	throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{
		LocalAgentPath localPath = getLocalPath();
		localPath.setCapabilityName(capabilityName);
		return getApplication().getAllForeignAgents(localPath);
	}
	
	/**
	 * @param string
	 * @return
	 */
	public IAgentProxy getAllForeignAgents(String capabilityName, String agentName) 
	throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{
		LocalAgentPath localPath = getLocalPath();
		localPath.setCapabilityName(capabilityName);
		localPath.setAgentName(agentName);
		return getApplication().getAllForeignAgents(localPath);
	}

	/**
	 * @param appName
	 * @param string
	 * @return
	 */
	public IAgentProxy getForeignAgent(String nodeName, String capabilityName) 
	throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{
		return getApplication().getAgent(new AgentPath(nodeName, getApplication().getName(), capabilityName, getName()));
	}

	/**
	 * @return Returns the logger.
	 */
	public Logger getLogger()
	{
		return logger;
	}
	
	public boolean isSystemApplication()
	{
		return getApplication().isSystemApplication();
	}
	
	public abstract boolean isIdle();
	
	public abstract long getQueueSize();

}
