/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node;

import java.io.Serializable;
import java.util.StringTokenizer;

import org.almendra.commons.util.StringUtility;
import org.almendra.janet.cas.agents.*;


public class AgentPath extends LocalAgentPath implements IPath, Serializable
{
	private static final long serialVersionUID = -7198677373876568758L;

	public static final String LOCAL_NODE_TAG = "LOCAL_NODE";  //$NON-NLS-1$
	
	protected String nodeName = LOCAL_NODE_TAG;

	/**
	 * Creates a new <code>AgentPath</code> object.
	 *  
	 */
	protected AgentPath()
	{
		super();
	}

	/**
	 * Creates a new <code>AgentPath</code> object.
	 *  
	 */
	public AgentPath(String nodeName, String appName, String capabilityName, String agentName)
		throws IllegalArgumentException
	{
		super();
		init(nodeName, appName, capabilityName, agentName);
	}

	/**
	 * Creates a new <code>AgentPath</code> object.
	 *  
	 */
	public AgentPath(String qualifiedAgentPath) 
	throws IllegalArgumentException
	{
		super();
		StringTokenizer tokenizer = new StringTokenizer(qualifiedAgentPath, "."); //$NON-NLS-1$
		if (!tokenizer.hasMoreTokens())
			throw new IllegalArgumentException(qualifiedAgentPath + " contains no node name"); //$NON-NLS-1$
		String nodeName = tokenizer.nextToken();
		if (!tokenizer.hasMoreTokens())
			throw new IllegalArgumentException(qualifiedAgentPath + " contains no application name"); //$NON-NLS-1$
		String appName = tokenizer.nextToken();
		if (!tokenizer.hasMoreTokens())
			throw new IllegalArgumentException(qualifiedAgentPath + " contains no capability name"); //$NON-NLS-1$
		String capName = tokenizer.nextToken();
		if (!tokenizer.hasMoreTokens())
			throw new IllegalArgumentException(qualifiedAgentPath + " contains no agent name"); //$NON-NLS-1$
		String agentName = tokenizer.nextToken();
		init(nodeName, appName, capName, agentName);
	}

	/**
	 * Creates a new <code>AgentPath</code> object.
	 * 
	 * @param agentPath
	 */
	public AgentPath(AgentPath agentPath) throws IllegalArgumentException
	{
		super();
		agentPath.checkValidity();
		init(agentPath);
	}

	/**
	 * Creates a new <code>AgentPath</code> object.
	 * 
	 * @param nodeName
	 * @param agentPath
	 */
	public AgentPath(String nodeName, AgentPath agentPath)
	{
		super();
		init(agentPath);
		this.nodeName = nodeName;
	}
	
	/**
	 * Creates a new <code>AgentPath</code> object.
	 * 
	 * @param nodeName
	 * @param agentPath
	 */
	public AgentPath(String nodeName, LocalAgentPath agentPath)
	{
		super(agentPath);
		this.nodeName = nodeName;
	}

	/**
	 * @param agentPath
	 */
	protected void init(AgentPath agentPath) throws IllegalArgumentException
	{
		agentPath.checkValidity();
		init(
			agentPath.getNodeName(),
			agentPath.getApplicationName(),
			agentPath.getCapabilityName(),
			agentPath.getAgentName());
	}

	protected void init(String nodeName, String appName, String capabilityName, String agentName)
		throws IllegalArgumentException
	{
		checkValidity(nodeName, appName, capabilityName, agentName);

		init(appName, capabilityName, agentName);
		this.nodeName = nodeName;
	}

	public void checkValidity() 
	throws IllegalArgumentException
	{
		if (!StringUtility.isValid(getNodeName()))
			throw new IllegalArgumentException("illegal node name " + getNodeName()); //$NON-NLS-1$
		super.checkValidity(getApplicationName(), getCapabilityName(), getAgentName());
	}

	protected void checkValidity(String nodeName, String appName, String capabilityName, String agentName)
	throws IllegalArgumentException
	{
		if (!StringUtility.isValid(nodeName))
			throw new IllegalArgumentException("illegal node name " + nodeName); //$NON-NLS-1$

		super.checkValidity(appName, capabilityName, agentName);
	}

	public String getQualifiedPathName()
	{
		return getNodeName() + "." + super.getQualifiedPathName(); //$NON-NLS-1$
	}

	public String getNodeName()
	{
		return nodeName;
	}

	/**
	 * @param compareLevel
	 * @return
	 */
	public String getElement(int compareLevel)
	{
		if (compareLevel == NoSuchAgentException.APPLICATION_LEVEL)
			return getApplicationName();
		if (compareLevel == NoSuchAgentException.CAPABILITY_LEVEL)
			return getCapabilityName();
		if (compareLevel == NoSuchAgentException.AGENT_LEVEL)
			return getAgentName();
		return null;
	}

	public static AgentPath getSystemPath(String nodeName)
	{
		return new AgentPath(
			nodeName,
			SystemApplication.NAME,
			SystemApplication.CORE_CAPABILITY_NAME,
			SystemAgent.NAME);
	}

	public AgentPath getReplyPath(String nodeName)
	{
		return new AgentPath(
			nodeName,
			applicationName,
			capabilityName,
			agentName);
	}
	
	/**
	 * @param str
	 */
	protected void appendDisplayPath(StringBuffer str)
	{
		str.append(getNodeName() + "->"); //$NON-NLS-1$
		super.appendDisplayPath(str);
	}

	/**
	 * @param string
	 */
	public void setNodeName(String nodeName)
	{
		this.nodeName = nodeName;
	}
	
	public void setLocal()
	{
		nodeName = LOCAL_NODE_TAG;
	}
	
	public boolean isLocalTag()
	{
		return LOCAL_NODE_TAG.equals(nodeName);
	}
	
	public LocalAgentPath toLocalAgentPath()
	{
		return new LocalAgentPath(this);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object arg)
	{
		if(this == arg)
			return true;
		if(arg instanceof AgentPath)
		{
			AgentPath agentPath = (AgentPath) arg;
			return sameAttributeValues(agentPath);
		}
		return false;
	}
	
	/**
	 * @param path
	 * @return
	 */
	protected boolean sameAttributeValues(AgentPath path)
	{
		if(!super.sameAttributeValues(path))			
			return false;
		return nodeName.equals(path.getNodeName());
	}
	
	public String getCapabilityPath()
	{
		LocalCapabilityPath path = new LocalCapabilityPath(getApplicationName(), getCapabilityName());
		return path.getQualifiedPathName();
	}
}
