/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.collection.StringList;
import org.almendra.commons.util.collection.TreeNode;
import org.almendra.commons.util.exception.AlreadyRegisteredException;
import org.almendra.commons.util.logging.ILoggable;


/**
 * @author Oliver Plohmann
 */
public class ApplicationRegistry implements ILoggable
{
	public static final String LOG_EVENT_REGISTERED_APPLICATION = "REGISTERED_APPLICATION"; //$NON-NLS-1$
	public static final String LOG_EVENT_DEREGISTERED_APPLICATION = "DEREGISTERED_APPLICATION"; //$NON-NLS-1$

	protected Map<String, AbstractApplication> applicationRegistry = new HashMap<String, AbstractApplication>();
	protected SystemApplication systemApplication = null;
	protected Node node = null;
	protected Logger logger = null;

	protected ApplicationRegistry()
	{
		super();
		init();
	}

	public ApplicationRegistry(Node node, Logger logger)
	{
		super();
		this.node = node;
		this.logger = Logger.getChildLogger(logger);
		init();
	}

	protected void init()
	{
		logger.registerForEvents(registeredLogEvents());
	}

	protected void log(String eventName, String appName)
	{
		if(logger.isSubscribed(eventName))
		{	
			LogEntry entry = logger.getNewLogEntry();
			entry.setEvent(eventName);
			entry.setArgument(appName);
			logger.info(entry);
		}	
	}
	
	public SystemApplication getSystemApplication()
	{
		return systemApplication;
	}

	public AbstractApplication registerSystemApplication(SystemApplication sysApp)
	{
		systemApplication = sysApp;
		log(LOG_EVENT_REGISTERED_APPLICATION, sysApp.getName());
		return sysApp;
	}

	public Application registerApplication(Application app) throws AlreadyRegisteredException, IllegalArgumentException
	{
		String name = app.getName();
		if (!StringUtility.isValid(name))
			throw new IllegalArgumentException("application must have a name!"); //$NON-NLS-1$

		synchronized (applicationRegistry)
		{
			if (applicationRegistry.containsKey(name))
			{
				throw new AlreadyRegisteredException(
						"application with name " //$NON-NLS-1$ 
						+ name + " already bound with application registry"); //$NON-NLS-1$
			}

			app.setAndInitLogger(logger);
			app.setNode(node);
			applicationRegistry.put(name, app);

			log(LOG_EVENT_REGISTERED_APPLICATION, app.getName());
		}

		return app;
	}

	public void deregisterApplication(String name) throws NoSuchElementException
	{
		synchronized (applicationRegistry)
		{
			AbstractApplication app = applicationRegistry.get(name);
			if (app == null)
			{
				throw new NoSuchElementException("no application named " //$NON-NLS-1$ 
				+name + " registered with application registry"); //$NON-NLS-1$
			}
			app.stop();
			applicationRegistry.remove(name);
			log(LOG_EVENT_DEREGISTERED_APPLICATION, app.getName());
		}
	}

	public AbstractApplication getApplication(String applicationName) throws NoSuchElementException
	{
		if (applicationName.equals(SystemApplication.NAME))
			return systemApplication;
		synchronized (applicationRegistry)
		{
			AbstractApplication app = applicationRegistry.get(applicationName);
			if (app == null)
				throw new NoSuchElementException("no such application " //$NON-NLS-1$ 
				+applicationName + "!"); //$NON-NLS-1$
			return app;
		}
	}
	
	public List<AbstractApplication> getAllApplications()
	{
		List<AbstractApplication> apps = new LinkedList<AbstractApplication>();
		apps.add(getSystemApplication());
		synchronized (applicationRegistry)
		{
			for(AbstractApplication app : applicationRegistry.values())
				apps.add(app);
		}
		return apps;
	}

	public boolean isRegistered(String applicationName)
	{
		synchronized (applicationRegistry)
		{
			return applicationRegistry.containsKey(applicationName);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.logging.ILoggable#registeredLogEvents()
	 */
	public List<String> registeredLogEvents()
	{
		List<String> events = new StringList();
		events.add(LOG_EVENT_REGISTERED_APPLICATION);
		events.add(LOG_EVENT_DEREGISTERED_APPLICATION);
		return events;
	}

	/**
	 * 
	 */
	public List<AgentPath> getAllAgentPaths()
	{
		synchronized (applicationRegistry)
		{
			List<AgentPath> agentPaths = new ArrayList<AgentPath>();
			for(AbstractApplication app : applicationRegistry.values())
				agentPaths.addAll(app.toForeignDescriptor());
			agentPaths.addAll(getSystemApplication().getNonCoreAgents());
			return agentPaths;
		}
	}
	
	public List<TreeNode> getTreeDescriptors()
	{
		List<TreeNode> appDescs = new LinkedList<TreeNode>();
		for(AbstractApplication app : getAllApplications())
			appDescs.add(app.toTreeDescriptor());
		return appDescs;
	}

}
