/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.janet.cas.node;

import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.almendra.commons.INamedObject;
import org.almendra.commons.events.Event;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.collection.CollectionUtility;
import org.almendra.commons.util.collection.ListUtility;
import org.almendra.commons.util.collection.StringList;
import org.almendra.commons.util.exception.AlreadyRegisteredException;
import org.almendra.commons.util.logging.ILoggable;
import org.almendra.janet.cas.agents.Agent;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.commands.RegisterAgentCommand;
import org.almendra.janet.cas.handlers.RegisteredAgentParams;
import org.almendra.janet.cas.schedulers.ApplicationScheduler;
import org.almendra.janet.cas.schedulers.SystemScheduler;
import org.almendra.janet.cas.scheduling.IInterpreter;


/**
 * 
 * @author Oliver Plohmann
 */
public class Capability implements INamedObject, ILoggable
{
	protected Map<String, IInterpreter> interpreters = new HashMap<String, IInterpreter>();
	protected AbstractApplication application = null;
	protected String name = null;
	protected String descriptorFilePathName = null;
	protected List<AgentDescriptor> agentDescriptors = new ArrayList<AgentDescriptor>();
	protected List<List<String>> agentsSharingScheduler = new ArrayList<List<String>>();
	protected int numberOfSchedulers = -1;
	protected String executeWhenAppStarted = null;
	protected List<String> registeredAgentNames = new StringList(); 

	private Capability()
	{
		super();
	}

	public Capability(AbstractApplication application, String capabilityName) 
	throws IllegalArgumentException
	{
		checkApplication(application);
		checkCapability(capabilityName);
		this.application = application;
		this.name = capabilityName;
		this.application.addCapability(this);
	}
	
	public Capability(AbstractApplication application, String capabilityName, String[] agentNames) 
	throws IllegalArgumentException
	{
		this(application, capabilityName);
		addAgentNames(ListUtility.toStringList(agentNames));
	}

	public Capability(AbstractApplication application, String capabilityName, String agentName) 
	throws IllegalArgumentException
	{
		this(application, capabilityName);
		addAgentNames(ListUtility.toStringListFromSingleObject(agentName));
	}

	protected void checkApplication(AbstractApplication application)
	throws IllegalArgumentException
	{
		if (application == null)
			throw new IllegalArgumentException("application must not be null!"); //$NON-NLS-1$
	}
	
	protected void checkCapability(String capabilityName)
	throws IllegalArgumentException
	{
		if(!StringUtility.isValid(capabilityName))
			throw new IllegalArgumentException("invalid capability name: " + StringUtility.print(capabilityName)); //$NON-NLS-1$		
	}
	
	protected void checkInterpreter(IInterpreter interpreter)
	throws IllegalArgumentException
	{
		if (interpreter == null)
			throw new IllegalArgumentException("interpreter must not be null!");	//$NON-NLS-1$
		for(String commandName : interpreter.commandNames())
		{
			if(!StringUtility.isValid(commandName))
				throw new IllegalArgumentException("invalid interpreter command name: " + StringUtility.print(commandName)); //$NON-NLS-1$
		}
	}
	
	public boolean checkValidity()
	throws IllegalArgumentException
	{
		checkApplication(application);
		checkCapability(getName());
		
		if(getAgentNames().isEmpty())
			throw new IllegalArgumentException("capability " + getQualifiedName() + " has no agents!"); //$NON-NLS-1$ //$NON-NLS-2$
		
		if(interpreters.isEmpty())
			throw new IllegalArgumentException("capability " + getQualifiedName() + " has no interpreters!"); //$NON-NLS-1$ //$NON-NLS-2$
		
		for(IInterpreter interpreter : getInterpreters())
			checkInterpreter(interpreter);

		return true;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.cas.node.INamedObject#getName()
	 */
	public String getName()
	{
		return name;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.logging.ILoggable#registeredLogEvents()
	 */
	public List<String> registeredLogEvents()
	{
		return null;
	}

	public void add(IInterpreter interpreter)
	throws IllegalArgumentException, DescriptorParseException
	{
		for(String commandName : interpreter.commandNames())
		{
			if(!StringUtility.isValid(commandName))
				throw new IllegalArgumentException("invalid command name: " + commandName + " for interpreter " + interpreter); //$NON-NLS-1$ //$NON-NLS-2$
		
			IInterpreter existingInterpreterWithSameName = interpreters.get(commandName);
			if(existingInterpreterWithSameName != null)
			{
				StringBuffer buffer = new StringBuffer();
				buffer.append("Duplicate interpreter definition: Command name " + commandName + " is use by interpreters "); //$NON-NLS-1$ //$NON-NLS-2$
				buffer.append(existingInterpreterWithSameName.getClass().getName() + " and " + interpreter.getClass().getName() + "!"); //$NON-NLS-1$ //$NON-NLS-2$
				throw new DescriptorParseException(buffer.toString());
			}
				
			interpreters.put(commandName, interpreter);
		}
	}
	
	public boolean hasInterpreters()
	{
		return !interpreters.isEmpty();
	}

	/**
	 * @return
	 */
	public IInterpreter getInterpreter(String qualifiedCommandName)
	{
		return interpreters.get(qualifiedCommandName);
	}
	
	public List<IInterpreter> getInterpreters()
	{
		return ListUtility.toList(interpreters.values());
	}
	
	public Node getNode()
	{
		return application.node;
	}

	/**
	 * @param list
	 */
	public void setSharedSchedulers(List<List<String>> sharedSchedulers)
	throws IllegalArgumentException
	{
		if(sharedSchedulers.isEmpty())
			return;
		List<List<String>> allObjects = new ArrayList<List<String>>();
		for(List<String> strings : sharedSchedulers)
			allObjects.add(ListUtility.toList(strings));
		Set<String> duplicates = CollectionUtility.duplicates(allObjects);
		if(!duplicates.isEmpty())
		{
			String msg = ""; //$NON-NLS-1$
			Iterator<String> it = duplicates.iterator();
			while (it.hasNext())
			{
				String agentName = it.next();
				msg += agentName;
				if(it.hasNext())
					msg += ", "; //$NON-NLS-1$
			}
			throw new IllegalArgumentException("agent(s) " + msg + " are assigned to more than one thread pool!"); //$NON-NLS-1$ //$NON-NLS-2$
		}
		agentsSharingScheduler = sharedSchedulers;
	}

	/**
	 * Answer the names of agents that use the same schedulerAnchor 
	 * as agentName. Return an ampty list if none.
	 * @param agentName
	 * @return
	 */
	public List<String> getAgentsSharingSchedulerWith(String agentName)
	{
		if(agentsSharingScheduler.isEmpty())
			return new StringList();
		for(List<String> agents : agentsSharingScheduler)
		{
			if(agents.contains(agentName))
			{
				// TODO - check whether removing object from a non-copied collection is correct here
				agents.remove(agentName);
				return agents;
			}							
		}
		return new StringList();
	}

	/**
	 * @return
	 */
	public String getQualifiedName()
	{
		return application.getName() + "." + getName(); //$NON-NLS-1$
	}

	/**
	 * @return
	 */
	public int getNumberOfSchedulers()
	{
		if(numberOfSchedulers == -1)
		{	
			if(agentsSharingScheduler.isEmpty() )
			{
				numberOfSchedulers = getAgentNames().size();
			}
			else
			{
				numberOfSchedulers = agentsSharingScheduler.size();
				for(String agentName : getAgentNames())
				{
					if(!isSchedulerShared(agentName))
						numberOfSchedulers++;
				}
			}
		}
		return numberOfSchedulers;
	}

	/**
	 * @param agentName
	 * @return
	 */
	protected boolean isSchedulerShared(String agentName)
	{
		for(List<String> agentNames : agentsSharingScheduler)
		{
			if(agentNames.contains(agentName))
				return true;
		}
		return false;
	}

	/**
	 * @return Returns the application.
	 */
	public AbstractApplication getApplication()
	{
		return application;
	}

	/**
	 * @return Returns the agentNames.
	 */
	public List<String> getAgentNames()
	{
		List<String> agentNames = new StringList();
		for(AgentDescriptor desc : agentDescriptors)
			agentNames.add(desc.getName());
		return agentNames;
	}

	/**
	 * @param list
	 */
	public void addInterpreters(List<IInterpreter> interpreters) throws IllegalArgumentException, DescriptorParseException
	{
		for(IInterpreter interpreter : interpreters)
			add(interpreter);
	}
	
	/**
	 * @return Returns the executeWhenAppStarted.
	 */
	public String getExecuteWhenAppStarted()
	{
		return executeWhenAppStarted;
	}

	/**
	 * @param executeWhenAppStarted The executeWhenAppStarted to set.
	 */
	public void setExecuteWhenAppStarted(String executeWhenAppStarted)
	{
		this.executeWhenAppStarted = executeWhenAppStarted;
	}

	/**
	 * @param list
	 */
	protected void addAgentNames(List<String> agentNames)
	{
		for(String agentName : agentNames)
		{
			if(!StringUtility.isValid(agentName))
				throw new IllegalArgumentException("invalid agent name: " + agentName + "!"); //$NON-NLS-1$ //$NON-NLS-2$
			getAgentDescriptors().add(new AgentDescriptor(agentName));
		}
	}
	/**
	 * @return Returns the agentDescriptors.
	 */
	public List<AgentDescriptor> getAgentDescriptors()
	{
		return agentDescriptors;
	}

	/**
	 * @param agentDescriptors The agentDescriptors to set.
	 */
	public void setAgentDescriptors(List<AgentDescriptor> agentDescriptors)
	{
		this.agentDescriptors = agentDescriptors;
	}
	
	public AgentPath getAgentPath(String agentName)
	throws IllegalArgumentException
	{
		if(!StringUtility.isValid(agentName))
			throw new IllegalArgumentException("invalid agent name: " + agentName); //$NON-NLS-1$
		return new AgentPath(
			getApplication().getNode().getName(),
			getApplication().getName(),
			getName(),
			agentName);
	}

	/**
	 * @return
	 */
	public String getAgentNamesString()
	{
		StringBuffer buffer = new StringBuffer();
		Iterator<String> it = getAgentNames().iterator();
		while (it.hasNext())
		{
			String agentName = it.next();
			buffer.append(agentName);
			if(it.hasNext())
				buffer.append(" | "); //$NON-NLS-1$
		}
		return buffer.toString();
	}
	
	public AgentPath registerAgent(String agentName)
	throws AlreadyRegisteredException, RemoteException
	{
		return registerAgent(agentName, false);
	}
	
	/**
	 * 
	 * @param agentName
	 * @param autoincrement If an agent with the name agentName already exists, return agentName with an appended incremented counter
	 * @return
	 * @throws AlreadyRegisteredException
	 * @throws RemoteException
	 */
	public synchronized AgentPath registerAgent(String agentName, boolean autoincrement)
	throws AlreadyRegisteredException, RemoteException
	{
		AgentPath agentPath = new AgentPath(getNode().getName(), getApplication().getName(), getName(), agentName);
		if(!autoincrement && registeredAgentNames.contains(agentName))
			throw new AlreadyRegisteredException(" agent " + agentPath.getQualifiedPathName() + " already registered!"); //$NON-NLS-1$ //$NON-NLS-2$		
		
		agentPath = getAgentPathToNewAgent(agentName);
		ApplicationScheduler scheduler = createApplicationScheduler();				
		agentPath = createAgent(agentPath.getAgentName(), scheduler);
				
		try
		{
			RegisterAgentCommand cmd = new RegisterAgentCommand(new AgentPath(getNode().getName(), agentPath)); 
			IAgentProxy agents = getNode().getSystemApplication().getSystemAgents(getApplication().getName());
			agents.accept(cmd);
		} 
		catch (InvalidCommandException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getApplication().getLogger().exception(e);	// logical error if this happens - do not handle
		}
		
		RegisteredAgentParams params = new RegisteredAgentParams(agentPath);
		Event event = new Event(RegisteredAgentParams.EVENT_NAME, params);
		getNode().getNodeEventRegistry().occured(event);

		return agentPath;
	}
	
	/**
	 * @param agentName
	 * @return
	 */
	protected AgentPath getAgentPathToNewAgent(String agentName)
	{
		if(!registeredAgentNames.contains(agentName))
			return new AgentPath(getNode().getName(), getApplication().getName(), getName(), agentName);
		
		List<String> matchingAgentNames = new StringList();
		for(String registeredAgentName : registeredAgentNames)
		{
			if(registeredAgentName.indexOf(agentName) != -1)
				matchingAgentNames.add(registeredAgentName);
		}
		List<String> integerSuffixes = new StringList();
		for(String currentAgentName : matchingAgentNames)
		{
			String integerSuffix = StringUtility.getIntegerSuffix(currentAgentName);
			if(integerSuffix.length() > 0)
				integerSuffixes.add(integerSuffix);			
		}
		String newSuffix = ""; //$NON-NLS-1$
		if(integerSuffixes.isEmpty())
		{
			newSuffix = "-0"; //$NON-NLS-1$
		}
		else
		{
			Collections.sort(integerSuffixes);
			String largestIntegerSuffix = integerSuffixes.get(integerSuffixes.size() - 1);
			newSuffix = String.valueOf((Integer.valueOf(largestIntegerSuffix).intValue()) + 1);
		}
		return new AgentPath(getNode().getName(), getApplication().getName(), getName(), agentName + "-" + newSuffix); //$NON-NLS-1$
		
	}

	public void deregisterAgent(String agentName)
	{
		// TODO - send DeregisterAgentCommand to all nodes that have the same app
		// TODO - wait asynchronously for all acks then stop the agent and remove the agent from local agentDispatcher
		boolean found = registeredAgentNames.remove(agentName);
		AssertUtility.assertTrue(found);
	}
	
	protected SystemScheduler createSubordinateSystemScheduler()
	{
		Node node = getNode();
		SystemScheduler scheduler =
			new SystemScheduler(
				node.getSystemArbitrator(),
				getApplication(), // TODO - unclean cast, cleanup design when more time
				this,
				node.getLogger(),
				true,
				SchedulerAnchor.SYSTEM_PRIORITY);
		return scheduler;
	}
	
	protected ApplicationScheduler createApplicationScheduler()
	{
		Node node = getNode();
		ApplicationScheduler scheduler =
			new ApplicationScheduler(
				node.getArbitrator(),
				getApplication(), // TODO - unclean cast, cleanup design when more time
				this,
				node.getLogger(),
				true,
				Thread.NORM_PRIORITY);
		return scheduler;
	}
	
	protected AgentPath createAgent(String agentName, ApplicationScheduler scheduler)
	{
		Node node = getNode();
		Agent agent = new Agent(node, this, scheduler, node.getLogger(), agentName);
		getNode().getLocalAgentDispatcher().add(agent);
		registeredAgentNames.add(agentName);
		return agent.getAgentPath();
	}
	
	public boolean hasAgent(String agentName)
	{
		return registeredAgentNames.contains(agentName);
	}

	/**
	 * @return Returns the registeredAgentNames.
	 */
	public List<String> getRegisteredAgentNames()
	{
		return registeredAgentNames;
	}
	/**
	 * @return Returns the descriptorFilePathName.
	 */
	public String getDescriptorFilePathName()
	{
		return descriptorFilePathName;
	}
	/**
	 * @param descriptorFilePathName The descriptorFilePathName to set.
	 */
	public void setDescriptorFilePathName(String descriptorFilePathName)
	{
		this.descriptorFilePathName = descriptorFilePathName;
	}
}
