/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node;

import java.io.Serializable;
import java.util.StringTokenizer;

import org.almendra.commons.util.StringUtility;
import org.almendra.janet.cas.agents.*;


/**
 * Defined the local path to an agent. Local means that the path can only be
 * resolved within the same node.
 */

public class LocalAgentPath extends LocalCapabilityPath implements IPath, Serializable
{
	private static final long serialVersionUID = -4806837309197866706L;
	protected String agentName = null;

	/**
	 * Creates a new <code>LocalAgentPath</code> object.
	 *  
	 */
	protected LocalAgentPath()
	{
		super();
	}

	/**
	 * Creates a new <code>LocalAgentPath</code> object.
	 *  
	 */
	public LocalAgentPath(LocalAgentPath path)
	{
		super();
		init(path.getApplicationName(), path.getCapabilityName(), path.getAgentName());
	}

	/**
	 * Creates a new <code>LocalAgentPath</code> object.
	 *  
	 */
	public LocalAgentPath(AgentPath path)
	{
		super();
		init(path.getApplicationName(), path.getCapabilityName(), path.getAgentName());
	}

	/**
	 * Creates a new <code>AgentPath</code> object.
	 *  
	 */
	public LocalAgentPath(String appName, String capabilityName, String agentName)
	{
		super();
		init(appName, capabilityName, agentName);
	}

	/**
	 * Creates a new <code>LocalAgentPath</code> object.
	 * 
	 * @param key
	 */
	public LocalAgentPath(String dottedPathString) throws IllegalArgumentException
	{
		super();
		StringTokenizer tokenizer = new StringTokenizer(dottedPathString, "."); //$NON-NLS-1$
		if (!tokenizer.hasMoreTokens())
			throw new IllegalArgumentException(dottedPathString + " contains no application name"); //$NON-NLS-1$
		String appName = tokenizer.nextToken();
		if (!tokenizer.hasMoreTokens())
			throw new IllegalArgumentException(dottedPathString + " contains no capability name"); //$NON-NLS-1$
		String capName = tokenizer.nextToken();
		if (!tokenizer.hasMoreTokens())
			throw new IllegalArgumentException(dottedPathString + " contains no agent name"); //$NON-NLS-1$
		String agentName = tokenizer.nextToken();
		init(appName, capName, agentName);
	}

	/**
	 * @param string
	 * @param string2
	 * @param string3
	 */
	protected void init(String appName, String capabilityName, String agentName) throws IllegalArgumentException
	{
		checkValidity(appName, capabilityName, agentName);
		init(appName, capabilityName);
		this.agentName = agentName;
	}

	public String getAgentName()
	{
		return agentName;
	}

	/**
	 * @return
	 */
	public String getLocalAgentName()
	{
		return super.getQualifiedPathName() + "." + getAgentName(); //$NON-NLS-1$
	}
	
	/**
	 * @return
	 */
	public String getQualifiedPathName()
	{
		return getLocalAgentName();
	}

	/**
	 * @param appName
	 * @param capabilityName
	 * @param agentName
	 */
	public void checkValidity(String appName, String capabilityName, String agentName)
	{
		if (!StringUtility.isValid(appName))
			throw new IllegalArgumentException("illegal application name " + appName); //$NON-NLS-1$

		if (!StringUtility.isValid(capabilityName))
			throw new IllegalArgumentException("illegal capability name " + capabilityName); //$NON-NLS-1$

		if (!StringUtility.isValid(agentName))
			throw new IllegalArgumentException("illegal agent name " + agentName); //$NON-NLS-1$

	}
	
	public void checkValidity() 
	throws IllegalArgumentException
	{
		checkValidity(getApplicationName(), getCapabilityName(), getAgentName());
	}

	/**
	 * @param path
	 * @return
	 */
	public int comparePath(LocalAgentPath path)
	{
		int compareLevel = NoSuchAgentException.APPLICATION_LEVEL;
		if (getApplicationName().equals(path.getApplicationName()))
			compareLevel = NoSuchAgentException.CAPABILITY_LEVEL;
		else
			return compareLevel;
		if (getCapabilityName().equals(path.getCapabilityName()))
			compareLevel = NoSuchAgentException.AGENT_LEVEL;
		else
			return compareLevel;
		if (getAgentName().equals(path.getAgentName()))
			compareLevel = NoSuchAgentException.NO_ERROR;
		return compareLevel;
	}

	public static LocalAgentPath getLocalSystemPath()
	{
		return new LocalAgentPath(
			SystemApplication.NAME, 
			SystemApplication.CORE_CAPABILITY_NAME, 
			SystemAgent.NAME);
	}
	
	public boolean isLocalTag()
	{
		return true;
	}
	/**
	 * @param agentName The agentName to set.
	 */
	public void setAgentName(String agentName)
	{
		this.agentName = agentName;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		StringBuffer str = new StringBuffer(super.toString());
		str.append("{"); //$NON-NLS-1$
		appendDisplayPath(str);
		str.append("}"); //$NON-NLS-1$
		return str.toString();
	}

	/**
	 * @param str
	 */
	protected void appendDisplayPath(StringBuffer str)
	{
		str.append(getApplicationName() + "->"); //$NON-NLS-1$
		str.append(getCapabilityName() + "->"); //$NON-NLS-1$
		str.append(getAgentName());
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object arg)
	{
		if(this == arg)
			return true;
		if(arg instanceof LocalAgentPath)
		{
			LocalAgentPath path = (LocalAgentPath) arg;
			return sameAttributeValues(path);
		}
		return false;
	}
	
	/**
	 * @param path
	 * @return
	 */
	protected boolean sameAttributeValues(LocalAgentPath path)
	{
		if(!super.sameAttributeValues(path))			
			return false;
		return agentName.equals(path.getAgentName());
	}
}
