/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.node;

import java.io.Serializable;

import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.xml.XMLUtility;
import org.almendra.janet.cas.scheduling.ICommandEnvelope;


/**
 * LogEntry class that adds additional Janet-specific attributes to the default
 * LogEntry class from org.almendra.commons.commons.util.logging.
 * 
 * @author Oliver Plohmann
 */

public class LogEntry extends org.almendra.commons.util.logging.LogEntry implements Cloneable, Serializable
{
	private static final long serialVersionUID = 8336985497891728111L;
	
	public static final String TAG_AGENT = "agent"; //$NON-NLS-1$
	public static final String TAG_CAPABILITY = "cap"; //$NON-NLS-1$
	public static final String TAG_APPLICATION = "app"; //$NON-NLS-1$
	public static final String TAG_SCHEDULER = "schedulerAnchor"; //$NON-NLS-1$
	public static final String TAG_OBJECT = "obj"; //$NON-NLS-1$
	public static final String TAG_NODE = "node"; //$NON-NLS-1$
	public static final String TAG_WORKSTATION = "workstation"; //$NON-NLS-1$
	public static final String TAG_COMMAND = "command"; //$NON-NLS-1$
	public static final String TAG_EXCEPTION_MSG = "exceptionMsg"; //$NON-NLS-1$
	public static final String TAG_EXCEPTION_STACK = "exceptionStack"; //$NON-NLS-1$
	
	protected String agentName = null;
	protected String applicationName = null;
	protected String capabilityName = null;
	protected String nodeName = null;
	protected String objectName = null;
	protected String schedulerName = null;
	protected String workstationName = null;
	protected String commandName = null;
	protected String exceptionMessage = null;
	protected String exceptionStack = null;
	protected Throwable exception = null;
	
	protected ICommandEnvelope commandEnvelope = null;

	/**
	 * Creates a new <code>LogEntry</code> object.
	 * 
	 * @param level
	 */
	public LogEntry(int level)
	{
		super(level);
	}

	/**
	 * Creates a new <code>LogEntry</code> object.
	 * 
	 * @param level
	 * @param argument
	 */
	public LogEntry(int level, String argument)
	{
		super(level, argument);
	}

	/**
	 * Creates a new <code>LogEntry</code> object.
	 * @param entry
	 */
	public LogEntry(LogEntry entry)
	{
		super();
		fillFromEntryTemplate(entry);
	}

	/**
	 * Return node name.
	 * 
	 * @return
	 */
	public String getNodeName()
	{
		return nodeName;
	}

	/**
	 * Return the object name.
	 * 
	 * @return
	 */
	public String getObjectName()
	{
		return objectName;
	}

	/**
	 * Return the schedulerAnchor name.
	 * 
	 * @return
	 */
	public String getSchedulerName()
	{
		return schedulerName;
	}

	/**
	 * Return workstation name.
	 * 
	 * @return
	 */
	public String getWorkstationName()
	{
		return workstationName;
	}

	/**
	 * Return capability name.
	 * 
	 * @return
	 */
	public String getCapabilityName()
	{
		return capabilityName;
	}

	/**
	 * Return application name.
	 * 
	 * @return
	 */
	public String getApplicationName()
	{
		return applicationName;
	}

	/**
	 * Return agent name.
	 * 
	 * @return
	 */
	public String getAgentName()
	{
		return agentName;
	}

	/**
	 * Set node name.
	 * 
	 * @return
	 */
	public void setNodeName(String nodeName)
	{
		this.nodeName = nodeName;
	}

	/**
	 * Set the object name.
	 * 
	 * @return
	 */
	public void setObjectName(String objectName)
	{
		this.objectName = objectName;
	}

	/**
	 * Set the schedulerAnchor name.
	 * 
	 * @return
	 */
	public void setSchedulerName(String schedulerName)
	{
		this.schedulerName = schedulerName;
	}

	/**
	 * Set the workstation name.
	 * 
	 * @return
	 */
	public void setWorkstationName(String workstationName)
	{
		this.workstationName = workstationName;
	}

	/**
	 * Set the capability name.
	 * 
	 * @return
	 */
	public void setCapabilityName(String capabilityName)
	{
		this.capabilityName = capabilityName;
	}

	/**
	 * Set the application name.
	 * 
	 * @return
	 */
	public void setApplicationName(String applicationName)
	{
		this.applicationName = applicationName;
	}

	/**
	 * Set the agent name.
	 * 
	 * @return
	 */
	public void setAgentName(String agentName)
	{
		this.agentName = agentName;
	}

	public String toLogString()
	{
		String separator = getLogStringSeparator();
		StringBuffer buffer = new StringBuffer(super.toLogString());
		buffer.append(separator);

		if (getAgentName() != null)
		{
			buffer.append(TAG_AGENT + "=" + StringUtility.print(getAgentName())); //$NON-NLS-1$
			buffer.append(separator);
		}
		if (getCapabilityName() != null)
		{
			buffer.append(TAG_CAPABILITY + "=" + StringUtility.print(getCapabilityName())); //$NON-NLS-1$
			buffer.append(separator);
		}
		if (getApplicationName() != null)
		{
			buffer.append(TAG_APPLICATION + "=" + StringUtility.print(getApplicationName())); //$NON-NLS-1$
			buffer.append(separator);
		}
		if (getSchedulerName() != null)
		{
			buffer.append(TAG_SCHEDULER + "=" + StringUtility.print(getSchedulerName())); //$NON-NLS-1$
			buffer.append(separator);
		}
		if (getObjectName() != null)
		{
			buffer.append(TAG_OBJECT + "=" + StringUtility.print(getObjectName())); //$NON-NLS-1$
			buffer.append(separator);
		}
				
		appendExceptionToLogString(buffer);
		
		if (getNodeName() != null)
		{
			buffer.append(TAG_NODE + "=" + StringUtility.print(getNodeName())); //$NON-NLS-1$
			buffer.append(separator);
		}
		if (getWorkstationName() != null)
		{
			buffer.append(TAG_WORKSTATION + "=" + StringUtility.print(getWorkstationName())); //$NON-NLS-1$ 
			buffer.append(separator);
		}

		String logString = buffer.toString();

		if (logString.length() == 0)
			return logString;

		if (logString.endsWith(separator))
			logString = logString.substring(0, logString.length() - separator.length());

		return logString;
	}

	/**
	 * @param buffer
	 */
	protected void appendCommandToLogString(StringBuffer buffer)
	{
		if(commandEnvelope == null)
			return;
		String commandName = StringUtility.reversePackageName(commandEnvelope.getCommand().getQualifiedName());
		buffer.append(TAG_COMMAND + "=\"" + commandName); //$NON-NLS-1$
		buffer.append("(id:" + commandEnvelope.getGlobalId().getValue() + ")\" "); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * @param exception2
	 */
	protected void appendExceptionToLogString(StringBuffer buffer)
	{
		if (exception != null)
		{
			appendCommandToLogString(buffer);
			boolean validMsg = false; // StringUtility.isValid(ex.getMessage());
			if(validMsg)
				buffer.append("\n    "); //$NON-NLS-1$
			buffer.append(TAG_EXCEPTION_MSG + "=" + StringUtility.print(exception.getMessage() + " ")); //$NON-NLS-1$ //$NON-NLS-2$
			if(validMsg)
				buffer.append("\n"); //$NON-NLS-1$
			java.io.StringWriter sw = new java.io.StringWriter(); 
			java.io.PrintWriter pw = new java.io.PrintWriter(sw);
			exception.printStackTrace(pw);
			String stackTrace = sw.toString();
			// if(StringUtility.isValid(stackTrace))
			//	buffer.append("\n    "); //$NON-NLS-1$
			buffer.append(TAG_EXCEPTION_STACK + "="); //$NON-NLS-1$
			if(StringUtility.isValid(stackTrace))
				buffer.append("\n"); //$NON-NLS-1$
			buffer.append(StringUtility.print(stackTrace) + " "); //$NON-NLS-1$
			pw.close();
			buffer.append(getLogStringSeparator());
		}
	}

	/**
	 * Returns a clone of this this LogEntry.
	 * 
	 * @return a clone of this LogEntry.
	 */
	public Object clone() throws CloneNotSupportedException
	{
		LogEntry entry = (LogEntry) super.clone();
		return entry;
	}

	/**
	 * Return a new <code>LogEntry</code> object with default values.
	 * 
	 * @return
	 */
	public static LogEntry getEntryTemplate(Logger logger)
	{
		return new LogEntry(logger.getDefaultValues());
	}
	
	/**
	 * Fill a <code>LogEntry</code> with default values.
	 * @param entry
	 */
	protected void fillFromEntryTemplate(LogEntry entry)
	{
		super.fillFromEntryTemplate(entry);
		setWorkstationName(entry.getWorkstationName());
		setNodeName(entry.getNodeName());
		setApplicationName(entry.getApplicationName());
		setCapabilityName(entry.getCapabilityName());
		setObjectName(entry.getObjectName());
		setAgentName(entry.getAgentName());
		setSchedulerName(entry.getSchedulerName());
		setCommandEnvelope(entry.getCommandEnvelope());
	}

	/**
	 * @return Returns the exception.
	 */
	public Throwable getException()
	{
		return exception;
	}

	/**
	 * @param exception The exception to set.
	 */
	public void setException(Throwable exception)
	{
		this.exception = exception;
	}

	/**
	 * @return Returns the commandEnvelope.
	 */
	public ICommandEnvelope getCommandEnvelope()
	{
		return commandEnvelope;
	}

	/**
	 * @param commandEnvelope The commandEnvelope to set.
	 */
	public void setCommandEnvelope(ICommandEnvelope commandEnvelope)
	{
		this.commandEnvelope = commandEnvelope;
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.commons.util.logging.LogEntry#fillFromEntryTemplate(java.lang.String, java.lang.String[])
	 */
	public void fillFromEntryTemplate(String entryString, String[] levelLabels)
	{
		super.fillFromEntryTemplate(entryString, levelLabels);		
		setAgentName(XMLUtility.getXMLValue(entryString, TAG_AGENT));
		setCapabilityName(XMLUtility.getXMLValue(entryString, TAG_CAPABILITY));
		setApplicationName(XMLUtility.getXMLValue(entryString, TAG_APPLICATION));
		setNodeName(XMLUtility.getXMLValue(entryString, TAG_NODE));
		setObjectName(XMLUtility.getXMLValue(entryString, TAG_OBJECT));
		setSchedulerName(XMLUtility.getXMLValue(entryString, TAG_SCHEDULER));
		setWorkstationName(XMLUtility.getXMLValue(entryString, TAG_WORKSTATION));
		setCommandName(XMLUtility.getXMLValue(entryString, TAG_COMMAND));
		setExceptionMessage(XMLUtility.getXMLValue(entryString, TAG_EXCEPTION_MSG, TAG_EXCEPTION_STACK));
		setExceptionStack(XMLUtility.getXMLValue(entryString, TAG_EXCEPTION_STACK, TAG_NODE));
	}

	/**
	 * @return Returns the commandName.
	 */
	public String getCommandName()
	{
		return commandName;
	}

	/**
	 * @param commandName The commandName to set.
	 */
	public void setCommandName(String commandName)
	{
		this.commandName = commandName;
	}

	/**
	 * @return Returns the exceptionMessage.
	 */
	public String getExceptionMessage()
	{
		return exceptionMessage;
	}

	/**
	 * @param exceptionMessage The exceptionMessage to set.
	 */
	public void setExceptionMessage(String exceptionMessage)
	{
		this.exceptionMessage = exceptionMessage;
	}

	/**
	 * @return Returns the exceptionStack.
	 */
	public String getExceptionStack()
	{
		return exceptionStack;
	}

	/**
	 * @param exceptionStack The exceptionStack to set.
	 */
	public void setExceptionStack(String exceptionStack)
	{
		this.exceptionStack = exceptionStack;
	}

	/**
	 * @return
	 */
	public String toExpandedXMLString()
	{
		return XMLUtility.toExpandedXMLString(getLogMessage());
	}

	/**
	 * 
	 */
	public void aboutToBeSent()
	{
		commandEnvelope = null;
	}

}
