/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package org.almendra.janet.cas.node;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.almendra.commons.events.Event;
import org.almendra.commons.threads.FutureResult;
import org.almendra.commons.threads.TimeoutException;
import org.almendra.commons.util.FileUtility;
import org.almendra.commons.util.StringUtility;
import org.almendra.commons.util.xml.XMLParseException;
import org.almendra.janet.cas.handlers.NodeStartupEventParams;
import org.almendra.janet.cas.spaces.IObjectSpace;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;


/**
 * @author Oliver
 *  
 */
public class NodeStarter
{
	public static final int NODE_STARTUP_TIMEOUT_PERIOD = 180000;

	protected Node node = null;
	protected NodeDescriptor nodeDescriptor = null;
	protected boolean success = true;
	
	protected NodeStarter()
	{
		super();
	}

	public static void main(String[] args)
	{
		if(args == null || args.length == 0)
		{
			System.out.println("Usage: java [-options] org.almendra.commons.janet.cas.node.NodeStarter <nodeDescriptorPathFileName.xml>"); //$NON-NLS-1$
			return;
		}
		
		String nodeDescriptorFilePathName = args[0];
	
		try
		{
			NodeStarter.start(nodeDescriptorFilePathName, NODE_STARTUP_TIMEOUT_PERIOD);
		}
		catch (DescriptorParseException e)
		{
			System.err.println(e.getMessage());
			e.printStackTrace();
		}
		catch (NodeStartupException e)
		{
			System.err.println(e.getMessage());
			e.printStackTrace();
		}
	}

	public static NodeAccessor start(String nodeDescriptorFilePathName) 
	throws DescriptorParseException, NodeStartupException
	{
		Node node = new NodeStarter().startIt(nodeDescriptorFilePathName, NODE_STARTUP_TIMEOUT_PERIOD);
		return new NodeAccessor(node);
	}

	public static NodeAccessor start(String nodeDescriptorFilePathName, int timeoutInMillisOrZero) 
	throws DescriptorParseException, NodeStartupException, IllegalArgumentException
	{
		if(timeoutInMillisOrZero < 0)
			throw new IllegalArgumentException("timeout period must be >= 0!"); //$NON-NLS-1$		
		Node node = new NodeStarter().startIt(nodeDescriptorFilePathName, timeoutInMillisOrZero);
		return new NodeAccessor(node);
	}

	protected void createAndStartNode() 
	throws DescriptorParseException, NodeStartupException
	{	
		node = createNode();
		node.start();
	}

	/**
	 * @return
	 * @throws DescriptorParseException
	 * @throws NodeStartupException
	 */
	protected Node createNode() throws DescriptorParseException, NodeStartupException
	{
		return new Node(nodeDescriptor);
	}

	protected void parse(String nodeDescriptorFilePathName) 
	throws DescriptorParseException
	{
		try
		{
			parseNodeDescriptor(nodeDescriptorFilePathName);
		}
		catch (Exception e)
		{
			// TODO - eventually fix this, but not too important
			// node.getLogger().exception(e); causes exception: logger at this state not yet installed
			StringBuffer msg = new StringBuffer();
			msg.append("Error parsing " + NodeDescriptor.FILE_NAME + ": " + e.getMessage()); //$NON-NLS-1$ //$NON-NLS-2$)
			msg.append("\n"); //$NON-NLS-1$
			msg.append("Node startup aborted!"); //$NON-NLS-1$
			throw new DescriptorParseException(msg.toString(), e);
		}
	}

	protected void parseNodeDescriptor(String nodeDescriptorFilePathName) 
	throws JDOMException, IOException, XMLParseException
	{
		if (nodeDescriptorFilePathName == null || nodeDescriptorFilePathName.length() == 0)
			throw new IOException("invalid node descriptor: " + StringUtility.print(nodeDescriptorFilePathName) + "!"); //$NON-NLS-1$ //$NON-NLS-2$
		
		File file = new File(nodeDescriptorFilePathName);		
		if (!file.exists())
		{			
			String absolutePath = file.getCanonicalPath();
			throw new IOException("Node descriptor " + nodeDescriptorFilePathName + " not found! Startup directory: " + absolutePath); //$NON-NLS-1$ //$NON-NLS-2$
		}
	
		SAXBuilder builder = new SAXBuilder();
		Document doc = builder.build(file);
	
		Element root = doc.getRootElement();
		nodeDescriptor = NodeDescriptor.create(root);
		
		Map<String, String> values = new HashMap<String, String>();
		values.put(NodeDescriptor.DESCRIPTOR_BASE_DIR, FileUtility.getDir(file));
		nodeDescriptor.visit(values);
	}

	protected void registerApplications() throws NodeStartupException
	{		
		registerSystemApplication();
		registerUserApplications();
	}

	protected void registerSystemApplication() throws NodeStartupException
	{
		try
		{
			node.registerSystemCapabilities();
		}
		catch (Exception e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			node.getLogger().exception(e);
			StringBuffer msg = NodeStartupException.getMessageHeader();
			msg.append(e.getMessage());
			throw new NodeStartupException(msg.toString(), e.getCause());
		}	
	}

	protected void registerUserApplications() throws NodeStartupException
	{
		IObjectSpace objectSpace = node.getSystemApplication().getObjectSpace();
		List<Application> startupApplications = (List<Application>) objectSpace.remove(NodeAnchor.REGISTRY_KEY_STARTUP_APPLICATIONS);
		
		if(startupApplications == null)
			return; // no aps defined in node descriptor
	
		for(Application app : startupApplications)
		{	
			try
			{
				node.getAccessor().registerApplication(app);
			}
			catch (Exception e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
				node.getLogger().exception(e);
				StringBuffer msg = NodeStartupException.getMessageHeader();
				msg.append(e.getMessage());
				throw new NodeStartupException(msg.toString(), e.getCause());
			}
		}
	
	}

	protected void sendNodeStartupEvent(int timeoutInMillisOrZero, boolean showGUI)
	throws NodeStartupException
	{
		FutureResult future = new FutureResult("waiting for node to startup"); //$NON-NLS-1$
		NodeStartupEventParams eventParams = new NodeStartupEventParams(
				future, 
				node.getSystemApplication().getCentralRegistryLocation(),
				node.getSystemApplication().getLocalRegistryLocation(),
				node.getSystemApplication().getClusterSpaceDescriptors(),
				node.getSystemApplication().getClusterEventRegistryDescriptors(),
				showGUI);
		
		Event event = new Event(NodeStartupEventParams.EVENT_NAME, eventParams);		
		node.getSystemApplication().getEventRegistry().occured(event);
		
		try
		{
			future.acquire(timeoutInMillisOrZero);
		}
		catch (IllegalArgumentException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			node.getLogger().exception(e);
		}
		catch (TimeoutException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			node.getLogger().exception(e);
		}
		
		if(future.hasException())
			throw new NodeStartupException(future.getException());
		
		node.getClusterImage().addLocalNode(node.getLocalAgentDispatcher(),	node.getNodeLocation());
	}

	/**
	 * @param args
	 */
	protected Node startIt(String descriptorRootDir, int timeoutInMillisOrZero) 
	throws DescriptorParseException, NodeStartupException
	{
		parse(descriptorRootDir);
		startNode(timeoutInMillisOrZero);
		return node;
	}

	protected void startNode(int timeoutInMillisOrZero)
	throws NodeStartupException, DescriptorParseException
	{
		createAndStartNode();
		sendNodeStartupEvent(timeoutInMillisOrZero, nodeDescriptor.isShowGUI());
		registerApplications();
	}

}
