/*
 * Created on 18.09.2003
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.schedulers;

/**
 * The Arbitrator makes sure that an application schedulerAnchor cannot cause
 * starvation of other application schedulers. Therefore, each application
 * schedulerAnchor receives a certain time slice of the CPU time only. There is
 * no time slicing if there is only one application schedulerAnchor present.
 * AbstractApplication schedulers that are not started, were stopped, or have no
 * commands to execute are not time sliced, either. All application schedulers
 * that are not time sliced are stored in <nonTimeSlicedSchedulers>.
 */

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.almendra.commons.threads.IExecutable;
import org.almendra.commons.threads.ThreadUtility;
import org.almendra.commons.util.AssertUtility;
import org.almendra.commons.util.logging.ILoggable;
import org.almendra.janet.cas.node.LogEntry;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.node.SchedulerAnchor;

/**
 * @author Oliver Plohmann
 */
public class Arbitrator<AbstractApplicationSchedulerType extends AbstractApplicationScheduler> extends AbstractScheduler implements ILoggable
{
    public static final String ARBITRATOR_NAME = "ARBITRATOR"; //$NON-NLS-1$

    public static final String LOG_EVENT_CHANGING_ACTIVE_SCHEDULER = "CHANGING_ACTIVE_SCHEDULER"; //$NON-NLS-1$

    public static final String LOGGER_OBJECT_NAME = "ARBITRATOR"; //$NON-NLS-1$

    protected AbstractApplicationScheduler currentScheduler = null;

    protected volatile int timeSliceInMillis = 500;

    protected Node node = null;

    protected SchedulerAnchor scheduler = null;

    protected List<AbstractApplicationScheduler> sleepingSchedulers = new LinkedList<AbstractApplicationScheduler>();

    protected List<AbstractApplicationScheduler> waitingSchedulers = new LinkedList<AbstractApplicationScheduler>();

    protected Lock producerConsumerLock = new ReentrantLock();

    protected volatile boolean isEnabled = true;

    protected Arbitrator()
    {
        super();
        init();
    }

    public Arbitrator(SchedulerAnchor scheduler, Logger logger, int priority)
    {
        super(new ConcurrentLinkedQueue<IExecutable>(), true, priority, logger);
        this.name = ARBITRATOR_NAME;
        this.node = scheduler.getNode();
        this.scheduler = scheduler;
        init();
    }

    protected IExecutable getNext()
    {
        // TODO - put unlock blocks into a try-finally block
        ApplicationScheduler appScheduler = (ApplicationScheduler) super.getNext();

        producerConsumerLock.lock();

        if (appScheduler.isIdle())
        {
            setSleeping(appScheduler);
            producerConsumerLock.unlock();
            return getNext();
        }

        if (!hasNext())
        {
            setWaiting(appScheduler);
            producerConsumerLock.unlock();
            return getNext();
        }

        add(appScheduler);
        producerConsumerLock.unlock();

        return appScheduler;

    }

    protected void setSleeping(AbstractApplicationScheduler arbitratedScheduler)
    {
        arbitratedScheduler.setSleeping();
        sleepingSchedulers.add(arbitratedScheduler);
    }

    protected void setWaiting(ApplicationScheduler appScheduler)
    {
        appScheduler.setWaiting();
        waitingSchedulers.add(appScheduler);
    }

    /**
     * @param i
     */
    public void setTimeSliceInMillis(int i)
    {
        timeSliceInMillis = i;
    }

    protected void commandAddedTo(AbstractApplicationScheduler arbitratedScheduler)
    {
        if (arbitratedScheduler.isRunning())
            return;

        if (!isEnabled)
        {
            if (!arbitratedScheduler.isRunning())
                start(arbitratedScheduler);
            return;
        }

        producerConsumerLock.lock();

        try
        {
            sleepingSchedulers.remove(arbitratedScheduler);

            if (available() > 1)
            {
                setScheduled(arbitratedScheduler);
                return;
            }

            if (waitingSchedulers.size() > 0)
            {
                setScheduled(arbitratedScheduler);
                setWaitingSchedulerScheduled(waitingSchedulers);
                return;
            }

            if (available() == 0)
            {
                arbitratedScheduler.setWaiting();
                waitingSchedulers.add(arbitratedScheduler);
                start(arbitratedScheduler);
                return;
            }
        }
        finally
        {
            producerConsumerLock.unlock();
        }
    }

    protected void setScheduled(AbstractApplicationScheduler arbitratedScheduler)
    {
        arbitratedScheduler.setScheduled();
        add(arbitratedScheduler);
    }

    protected void setWaitingSchedulerScheduled(List arbitratedSchedulers)
    {
        while (!arbitratedSchedulers.isEmpty())
        {
            setScheduled((ApplicationScheduler) arbitratedSchedulers.remove(0));
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.almendra.commons.cas.node.scheduling.ActiveSchedulingObject#execute(java.lang.Object)
     */
    protected void execute(IExecutable executable)
    {
        if (executable instanceof AbstractApplicationScheduler)
        {
            if (currentScheduler != null)
                currentScheduler.setPriority(SchedulerAnchor.REDUCED_APPLICATION_PRIORITY);
            AbstractApplicationScheduler scheduler = (AbstractApplicationScheduler) executable;
            if (scheduler != null)
            {
                start(scheduler);
                sleep(scheduler);
            }
            currentScheduler = scheduler;
        }
    }

    protected void start(AbstractApplicationScheduler arbitratedScheduler)
    {
        int priority = arbitratedScheduler.getPriority();
        priority = Math.min(priority, arbitratedScheduler.getMaxPriority());
        arbitratedScheduler.setPriority(priority);
        logChangedActiveScheduler(currentScheduler, arbitratedScheduler);
        arbitratedScheduler.start();
    }

    protected void sleep(AbstractApplicationScheduler arbitratedScheduler)
    {
        int share = arbitratedScheduler.getNumberOfSchedulersInSameApplication();
        AssertUtility.assertTrue(share > 0);
        ThreadUtility.sleep(timeSliceInMillis / share);
    }

    /**
     * @param schedulerAnchor
     */
    protected void logChangedActiveScheduler(CommandScheduler previouslyActiveScheduler,
            CommandScheduler currentlyActiveScheduler)
    {
        if (logger.isSubscribed(LOG_EVENT_CHANGING_ACTIVE_SCHEDULER))
        {
            LogEntry entry = logger.getNewLogEntry();
            entry.setEvent(LOG_EVENT_CHANGING_ACTIVE_SCHEDULER);
            String argument = "prevActive:" + previouslyActiveScheduler + " | " + "currentActive:" + currentlyActiveScheduler; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
            entry.setArgument(argument);
            logger.info(entry);
        }
    }

    public void stop()
    {
        super.stop();
        while (queue.size() > 0)
        {
            AbstractApplicationScheduler scheduler = (AbstractApplicationScheduler) queue.poll();
            scheduler.stop();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.almendra.commons.cas.node.scheduling.ActiveObject#isDaemon()
     */
    public boolean isDaemon()
    {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.almendra.commons.commons.util.logging.ILoggable#registeredLogEvents()
     */
    public List<String> registeredLogEvents()
    {
        List<String> events = super.registeredLogEvents();
        events.add(LOG_EVENT_CHANGING_ACTIVE_SCHEDULER);
        return events;
    }

    protected void init()
    {
        if (logger != null)
        {
            logger.registerForEvents(registeredLogEvents());
        }
    }

    /**
     * @param appScheduler
     */
    public void created(AbstractApplicationScheduler arbitratedScheduler)
    {
        AssertUtility.assertTrue(arbitratedScheduler.available() == 0);
        arbitratedScheduler.setNew();
        sleepingSchedulers.add(arbitratedScheduler);
    }

    /**
     * @param appScheduler
     */
    public void removeScheduler(ApplicationScheduler appScheduler)
    {
        if (appScheduler.isActive())
        {
            remove(appScheduler);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.almendra.commons.commons.threads.ActiveSchedulingObject#isIdle()
     */
    public boolean isIdle()
    {
        return (waitingSchedulers.isEmpty() && super.isIdle());
    }

    /**
     * @param appScheduler
     */
    public void isIdle(AbstractApplicationScheduler arbitratedScheduler)
    {
        producerConsumerLock.lock();
        try
        {
            if (arbitratedScheduler.isWaiting())
            {
                waitingSchedulers.remove(arbitratedScheduler);
                setSleeping(arbitratedScheduler);
            }
        }
        finally
        {
            producerConsumerLock.unlock();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.almendra.commons.janet.cas.logger.schedulers.AbstractScheduler#setLogger(org.almendra.commons.janet.cas.logger.Logger)
     */
    protected void setLogger(Logger logger)
    {
        super.setLogger(logger);
        this.logger.getDefaultValues().setSchedulerName(LOGGER_OBJECT_NAME);
    }

    /**
     * @param b
     */
    public void start(boolean isEnabled)
    {
        this.isEnabled = isEnabled;
        if (isEnabled)
            start();
    }

}
