/*
 * Created on 18.09.2003
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.schedulers;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Queue;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.almendra.janet.cas.scheduling.*;

public class PriorityQueue<PrioritizedObjectType extends IPrioritizedObject> implements Queue<PrioritizedObjectType>
{
	protected List<List<PrioritizedObjectType>> queues = null;
	protected int size = 0;
	protected ReadWriteLock lock = new ReentrantReadWriteLock(true);

	public PriorityQueue()
	{
		super();
		init(0);
	}

	public PriorityQueue(int numQueues)
	{
        super();
		init(numQueues);
	}

    protected boolean isValid(int priority)
    {
        if (priority < 0 || priority > queues.size())
        {
            System.err.println("invalid parameter in " + getClass().getName() + ".add(...)\n" + //$NON-NLS-1$ //$NON-NLS-2$
                    "- value of priority out of range: " + priority); //$NON-NLS-1$     
            return false;
        }
        return true;
    }
    
	public boolean add(PrioritizedObjectType prioritizedObject)
	{
		int priority = prioritizedObject.getPriority();
        if(!isValid(priority))
            return false;	
		
		lock.writeLock().lock();
		try
		{
			List<PrioritizedObjectType> queue = queues.get(priority);
			queue.add(prioritizedObject);
			size++;
			return true;
		}
		finally
		{
			lock.writeLock().unlock();
		}
	}

	protected void init(int numQueues)
	{
		queues = new LinkedList<List<PrioritizedObjectType>>();
		for (int i = 0; i < numQueues; i++)
            queues.add(new LinkedList<PrioritizedObjectType>());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.scheduling.SharedQueue#isEmpty()
	 */
	public boolean isEmpty()
	{
	    lock.readLock().lock();
		try
		{			
			return size == 0;
		}
		finally
		{
			lock.readLock().unlock();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.scheduling.SharedQueue#size()
	 */
	public int size()
	{
	    lock.readLock().lock();
		try
		{
			return size;
		}
		finally
		{
			lock.readLock().unlock();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.almendra.commons.commons.util.collection.IQueue#iterator()
	 */
	public Iterator<PrioritizedObjectType> iterator()
	{
	    lock.readLock().lock();
		try
		{			
			LinkedList<PrioritizedObjectType> list = new LinkedList<PrioritizedObjectType>();
			for (int i = 0; i < queues.size(); i++)
			{
				List<PrioritizedObjectType> commands = queues.get(i);
				list.addAll(commands);
			}
			return list.iterator();
		}
		finally
		{
			lock.readLock().unlock();
		}
	}

    public boolean offer(PrioritizedObjectType object)
    {
        return add(object);
    }

    public PrioritizedObjectType poll()
    {
        lock.writeLock().lock();
        try
        {           
            for (int i = 0; i < queues.size(); i++)
            {
                List<PrioritizedObjectType> queue = queues.get(i);
                if (!queue.isEmpty())
                {
                    size--;
                    return queue.remove(0);
                }
            }
            return null;
        }
        finally
        {
            lock.writeLock().unlock();
        }
    }

    public PrioritizedObjectType remove()
    {
        PrioritizedObjectType prioritizedObject = poll();
        if(prioritizedObject == null)
            throw new NoSuchElementException("queue is empty"); //$NON-NLS-1$
        return prioritizedObject;
    }

    public PrioritizedObjectType peek()
    {
        lock.writeLock().lock();
        try
        {           
            for (int i = 0; i < queues.size(); i++)
            {
                List<PrioritizedObjectType> queue = queues.get(i);
                if (!queue.isEmpty())
                    return queue.get(0);
            }
            return null;
        }
        finally
        {
            lock.writeLock().unlock();
        }
    }

    public PrioritizedObjectType element()
    {
        PrioritizedObjectType prioritizedObject = peek();
        if(prioritizedObject == null)
            throw new NoSuchElementException("no such element " + prioritizedObject); //$NON-NLS-1$
        return prioritizedObject;
    }

    public boolean contains(Object object)
    {
        lock.writeLock().lock();
        try
        {
            if(object instanceof IPrioritizedObject)
            {
                IPrioritizedObject prioritizedObject = (IPrioritizedObject) object;
                int priority = prioritizedObject.getPriority(); 
                if(!isValid(priority))
                    return false;
                List<PrioritizedObjectType> queue = queues.get(priority);
                return queue.contains(prioritizedObject);
            }
            return false;
        }        
        finally
        {
            lock.writeLock().unlock();
        }
    }

    public Object[] toArray()
    {
        // TODO - implement later
        return null;
    }

    public <PrioritizedObjectType> PrioritizedObjectType[] toArray(PrioritizedObjectType[] a)
    {
        // TODO - implement later
        return null;
    }

    public boolean remove(Object object)
    {
        lock.writeLock().lock();
        try
        {
            if(object instanceof IPrioritizedObject)
            {
                IPrioritizedObject prioritizedObject = (IPrioritizedObject) object;
                int priority = prioritizedObject.getPriority(); 
                if(!isValid(priority))
                    return false;
                List<PrioritizedObjectType> queue = queues.get(priority);
                boolean changed = queue.remove(prioritizedObject);
                if(!changed)
                    return false;
                size--;
                return true;
            }
            return false;
        }        
        finally
        {
            lock.writeLock().unlock();
        }
    }

    public boolean containsAll(Collection<?> collection)
    {
        for(Object object : collection)
        {
            if(!contains(object))
                return false;
        }
        return true;
    }

    public boolean addAll(Collection<? extends PrioritizedObjectType> collection)
    {
        boolean changed = false;
        for(PrioritizedObjectType prioritizedObject : collection)
            changed |= add(prioritizedObject);
        return changed;
    }

    public boolean removeAll(Collection<?> collection)
    {
        boolean changed = false;
        for(Object object : collection)
            changed |= remove(object);
        return changed;
    }

    public boolean retainAll(Collection<?> collection)
    {
        boolean changed = false;
        for(Object object : collection)
        {
            if(!contains(object))
                changed |= remove(object);
        }
        return changed;
    }

    public void clear()
    {
        lock.writeLock().lock();
        try
        {
            if(queues == null)
                return;
            init(queues.size());
        }        
        finally
        {
            lock.writeLock().unlock();
        }
    }

}
