/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source code or
 * in binary form must reproduce the above copyright notice, and the disclaimer
 * that follows.
 * 
 * In addition, you are required that you include in the end-user documentation
 * provided with the redistribution and/or in the software itself an
 * acknowledgement equivalent to the following: "This product is based on the
 * Janet software developed by the JANET Project
 * (http://www.almendra.org/janet/)."
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE JDOM
 * AUTHORS OR THE PROJECT CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package org.almendra.janet.cas.scheduling;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.NoSuchElementException;

import org.almendra.commons.util.exception.InsufficientPrivilegeException;
import org.almendra.commons.util.exception.InvalidMessageException;
import org.almendra.janet.cas.agents.IAgent;
import org.almendra.janet.cas.agents.IAgentProxy;
import org.almendra.janet.cas.events.ILocalClusterEventRegistry;
import org.almendra.janet.cas.events.ILocalEventRegistry;
import org.almendra.janet.cas.node.AbstractNodeAnchor;
import org.almendra.janet.cas.node.AgentPath;
import org.almendra.janet.cas.node.ApplicationAccessor;
import org.almendra.janet.cas.node.Capability;
import org.almendra.janet.cas.node.IllegalAgentAccessException;
import org.almendra.janet.cas.node.LocalAgentPath;
import org.almendra.janet.cas.node.Logger;
import org.almendra.janet.cas.node.NoSuchAgentException;
import org.almendra.janet.cas.node.Node;
import org.almendra.janet.cas.node.NodeAccessor;
import org.almendra.janet.cas.node.NodeAnchor;
import org.almendra.janet.cas.node.NodeName;
import org.almendra.janet.cas.node.SystemApplication;
import org.almendra.janet.cas.node.gui.NodeViewTitleAccessor;
import org.almendra.janet.cas.spaces.IObjectSpace;


public class CommandAccessor
{
	public static final String LABEL = "CommandAccessor"; //$NON-NLS-1$
	
	protected ICommandEnvelope commandEnvelope = null;

	/**
	 * Creates a new <code>CommandAccessor</code> object.
	 */
	protected CommandAccessor()
	{
		super();
	}

	/**
	 * Creates a new <code>CommandAccessor</code> object.
	 * 
	 * @param commandEnvelope
	 */
	public CommandAccessor(CommandAccessor cmdAccessor)
	{
		super();
		this.commandEnvelope = new CommandEnvelope(cmdAccessor.commandEnvelope);
	}
	
	/**
	 * Creates a new <code>CommandAccessor</code> object.
	 * 
	 * @param commandEnvelope
	 */
	public CommandAccessor(ICommandEnvelope commandEnvelope)
	{
		super();
		this.commandEnvelope = commandEnvelope;
	}

	/**
	 * Creates a new <code>CommandAccessor</code> object.
	 */
	public CommandAccessor(CommandAccessor cmdAccessor, ICommand command)
	{
		this(cmdAccessor);
		commandEnvelope.setCommand(command);
	}

	/**
	 * @return
	 */
	public ICommand getCommand()
	{
		return commandEnvelope.getCommand();
	}

	/**
	 * @return
	 */
	public AbstractNodeAnchor getNodeAnchor() throws InsufficientPrivilegeException
	{
		if (!commandEnvelope.getCurrentAgent().getApplication().isSystemApplication())
			throw new InsufficientPrivilegeException("node anchor can only be accessed by the system agent"); //$NON-NLS-1$
		return (AbstractNodeAnchor) getObjectSpace().get(NodeAnchor.LABEL);
	}

	/**
	 * @return
	 */
	public IObjectSpace getObjectSpace()
	{
		return getApplication().getObjectSpace();
	}

	/**
	 * @param replyObject
	 */
	public void setReply(Serializable replyObject) throws InvalidMessageException
	{
		commandEnvelope.setResponse(replyObject);
	}

	/**
	 * @return
	 */
	public String getApplicationName()
	{
		return commandEnvelope.getApplicationName();
	}

	/**
	 * @return
	 */
	public ApplicationAccessor getApplication()
	{
		return commandEnvelope.getCurrentAgent().getApplication().getAccessor();
	}

	public Capability getCapability()
	{
		return getApplication().getCapability(getCapabilityName());
	}

	/**
	 * @return
	 */
	public String getCapabilityName()
	{
		return commandEnvelope.getCapabilityName();
	}

	/**
	 * @return
	 */
	public CommandHistory getCommandHistory()
	{
		return commandEnvelope.getHistory();
	}

	/**
	 * @return
	 */
	public String getLocalizedDestinationAgentName()
	{
		return commandEnvelope.getLocalizedDestinationAgentName();
	}

	/**
	 * @return
	 */
	public String getQualifiedDestinationAgentName()
	{
		return commandEnvelope.getQualifiedDestinationAgentName();
	}

	/**
	 * @return @throws
	 *         InsufficientPrivilegeException
	 */
	public Node getNode() throws InsufficientPrivilegeException
	{
		return commandEnvelope.getNode();
	}

	/**
	 * @return @throws
	 *         InsufficientPrivilegeException
	 */
	public String getNodeName()
	{
		return commandEnvelope.getNodeName();
	}
	
	/**
	 * @return @throws
	 *         InsufficientPrivilegeException
	 */
	public NodeName getNodeNameObject()
	{
		return commandEnvelope.getNodeNameObject();
	}
	
	public String getCommandId()
	{
		return commandEnvelope.getGlobalId().getValue();
	}
	
	/**
	 * @return @throws
	 *         InsufficientPrivilegeException
	 */
	public ICommandEnvelope getCommandEnvelope() throws InsufficientPrivilegeException
	{
		if(getApplication().isSystemApplication())
			return commandEnvelope;
		throw new InsufficientPrivilegeException("privilege only granted to system agent"); //$NON-NLS-1$
	}
	
	/**
	 * @return @throws
	 *         InsufficientPrivilegeException
	 */
	public NodeAccessor getNodeAccessor()
	{
		return commandEnvelope.getNodeAccessor();
	}

	/**
	 *  
	 */
	public SystemApplication getSystemApplication() throws InsufficientPrivilegeException
	{
		return commandEnvelope.getCurrentAgent().getSystemApplication();
	}
	
	/**
	 *  
	 */
	public SystemApplication getSystemApplicationIfPrivilege()
	{
		try
		{
			return commandEnvelope.getCurrentAgent().getSystemApplication();
		}
		catch (InsufficientPrivilegeException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * @param nodeName
	 * @return
	 */
	public IAgentProxy getSystemAgent(String nodeName)
		throws NoSuchAgentException, IllegalAgentAccessException, InsufficientPrivilegeException
	{
		return getSystemApplication().getSystemAgent(nodeName);
	}

	/**
	 * @param nodeName
	 * @return
	 */
	public IAgentProxy getSystemAgent() throws NoSuchAgentException, IllegalAgentAccessException, InsufficientPrivilegeException
	{
		return getSystemApplication().getSystemAgent();
	}

	/**
	 * @param nodeName
	 * @return
	 */
	public IAgentProxy getAgent(AgentPath agentPath) throws NoSuchAgentException, IllegalAgentAccessException
	{
		return getApplication().getAgent(agentPath);
	}
	
	/**
	 * @param nodeName
	 * @return
	 */
	public IAgentProxy getCrossAgent(AgentPath agentPath) 
	throws NoSuchAgentException, IllegalAgentAccessException, InsufficientPrivilegeException
	{
		return commandEnvelope.getCrossAgent(agentPath);
	}

	/**
	 * @param ex
	 */
	public void log(Exception ex)
	{
		commandEnvelope.log(ex);
	}

	/**
	 * @return
	 */
	public Logger getLogger()
	{
		return commandEnvelope.getLogger();
	}

	/**
	 * @return
	 */
	public IAgent getCurrentAgent()
	{
		return commandEnvelope.getCurrentAgent();
	}

	/**
	 * @return @throws
	 *         NoSuchAgentException
	 * @throws IllegalAgentAccessException
	 */
	public IAgentProxy getCurrentAgentProxy()
	{
		return commandEnvelope.getCurrentAgentProxy();
	}

	/**
	 * @return
	 */
	public ILocalEventRegistry getNodeEventRegistry()
	{
		return commandEnvelope.getNodeEventRegistry();
	}

	/**
	 * @return
	 */
	public IObjectSpace getNodeObjectSpace()
	{
		return commandEnvelope.getNodeObjectSpace();
	}
	
	/**
	 * @return
	 */
	public IObjectSpace getApplicationObjectSpace()
	{
		return commandEnvelope.getApplicationObjectSpace();
	}

	public NodeAnchor getNodeAnchorIfPrivilege()
	{
		try
		{
			return (NodeAnchor) getNodeAnchor();
		}
		catch (InsufficientPrivilegeException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
		}
		return null;
	}

	public AbstractNodeAnchor getAbstractNodeAnchorIfPrivilege()
	{
		try
		{
			return getNodeAnchor();
		}
		catch (InsufficientPrivilegeException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
		}
		return null;
	}

	public Node getNodeIfPrivilege()
	{
		try
		{
			return getNode();
		}
		catch (InsufficientPrivilegeException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
		}
		return null;
	}
	
	public ICommandEnvelope getCommandEnvelopeIfPrivilege()
	{
		try
		{
			return getCommandEnvelope();
		}
		catch (InsufficientPrivilegeException e)
		{
			System.out.println(e.getMessage());
			e.printStackTrace();
			getLogger().exception(e);
		}
		return null;
	}

	/**
	 * @return
	 */
	public AgentPath getAgentPath()
	{
		return commandEnvelope.getAgentPath();
	}
	
	public LocalAgentPath getLocalAgentPath()
	{
		return commandEnvelope.getLocalAgentPath();
	}

	/**
	 * @return
	 */
	public IAgentProxy getAgentProxy()
	{
		return commandEnvelope.getAgentProxy();
	}

	/**
	 * @param capabilityName
	 * @return
	 * @throws IllegalArgumentException
	 * @throws NoSuchAgentException
	 * @throws IllegalAgentAccessException
	 */
	public IAgentProxy getAllForeignAgents(String capabilityName)
		throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{
		return commandEnvelope.getAllForeignAgents(capabilityName);
	}

	/**
	 * @param nodeName
	 * @param capabilityName
	 * @return
	 * @throws IllegalArgumentException
	 * @throws NoSuchAgentException
	 * @throws IllegalAgentAccessException
	 */
	public IAgentProxy getForeignAgent(String nodeName, String capabilityName)
		throws IllegalArgumentException, NoSuchAgentException, IllegalAgentAccessException
	{
		return commandEnvelope.getForeignAgent(nodeName, capabilityName);
	}

	/**
	 * @return
	 */
	public Object getOriginNodeName()
	{
		return commandEnvelope.getOriginNodeName();
	}

	/**
	 * @param name
	 * @return
	 * @throws NoSuchElementException
	 * @throws RemoteException
	 */
	public ILocalClusterEventRegistry getClusterEventRegistry(String name) 
	throws NoSuchElementException, RemoteException 
	{
		return commandEnvelope.getClusterEventRegistry(name);
	}
	/**
	 * @return
	 */
	public ILocalEventRegistry getApplicationEventRegistry()
	{
		return commandEnvelope.getApplicationEventRegistry();
	}

	/**
	 * @return
	 */
	public String getQualifiedCapacityName()
	{
		return commandEnvelope.getQualifiedCapabilityName();
	}

	/**
	 * @param string
	 */
	public void setNodeLabel(String label)
	{
		AbstractNodeAnchor nodeAnchor = (AbstractNodeAnchor) getObjectSpace().get(NodeAnchor.LABEL);
		if(nodeAnchor == null)
			return;
		nodeAnchor.setLabel(label);
	}
	
	public NodeViewTitleAccessor nodeViewTitleAccessorIfPrivilege()
	{
		SystemApplication app = getSystemApplicationIfPrivilege();
		if(app == null)
			return null;
		return (NodeViewTitleAccessor) getObjectSpace().get(NodeViewTitleAccessor.LABEL);
	}

	/**
	 * @return
	 */
	public boolean isSuspended()
	{
		return commandEnvelope.isSuspended();
	}
}
