/*
 * 
 * Copyright (C) 2004, 2005 Oliver Plohmann
 * All rights reserved.
 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that redistributions of source    
 * code or in binary form must reproduce the above copyright notice, 
 * and the disclaimer that follows.

 * In addition, you are required that you include in the  end-user 
 * documentation provided with the redistribution and/or in the 
 * software itself an acknowledgement equivalent to the following:
 * "This product is based on the Janet software developed by the 
 * JANET Project (http://www.almendra.org/janet/)."

 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE JANET AUTHORS OR THE PROJECT
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
package org.almendra.janet.cas.spaces;

import java.util.HashMap;
import java.util.Map;


public class ObjectSpace extends AbstractObjectSpace implements IObjectSpace
{
	/**
	 * The map to store objects stored by the object space.  
	 */
	protected Map<String, Object> objects = new HashMap<String, Object>();
	
	/**
	 * The map to store <code>IAttacher</code> objects by the key of the object they are associated with.
	 */
	protected Map<String, IAttacher> attachers = new HashMap<String, IAttacher>();
	
	/**
	 * The map to store <code>IAttacher</code> objects by the key of the object they are associated with.
	 */
	protected Map<String, IDetacher> detachers = new HashMap<String, IDetacher>();
	/**
	 * Creates a new <code>ObjectSpace</code> object.
	 */
	
	public ObjectSpace()
	{
		super();
	}

	/*
	 * (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#removeAttacher(java.lang.String)
	 */
	public IAttacher removeAttacher(String key)
	{
		accessProtection.lock();
		try
		{
		    return attachers.remove(key);
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#removeDetacher(java.lang.String)
	 */
	public IDetacher removeDetacher(String key)
	{
		accessProtection.lock();
		try
		{
		    return detachers.remove(key);
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#get(java.lang.String)
	 */
	public Object get(String key)
	{
		accessProtection.lock();
		try
		{
		    return objects.get(key);
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#remove(java.lang.String)
	 */
	public Object remove(String key)
	{
		accessProtection.lock();
		try
		{
		    return objects.remove(key);
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#detach(java.lang.String)
	 */
	public Object detach(String key) throws NoInstalledDetacher
	{
		accessProtection.lock();
		try
		{
		    Object object = objects.get(key);
			IDetacher detacher = detachers.get(key);
			if (detacher == null)
				throwNoInstalledDetacher(key);
			return detacher.detach(object);
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#put(java.lang.String, java.lang.Object)
	 */
	public void put(String key, Object object)
	{
		accessProtection.lock();
		try
		{
		    objects.put(key, object);
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/*
	 * (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#installAttacher(java.lang.String, org.almendra.commons.janet.cas.spaces.IAttacher)
	 */
	public void installAttacher(String key, IAttacher attacher)
	{
		accessProtection.lock();
		try
		{
		    attachers.put(key, attacher);
		}
		finally
		{
		    accessProtection.unlock();
		}		
	}

	/*
	 * (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#installDetacher(java.lang.String, org.almendra.commons.janet.cas.spaces.ISerializableDetacher)
	 */
	public void installDetacher(String key, IDetacher detacher)
	{
		accessProtection.lock();
		try
		{
		    detachers.put(key, detacher);
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#attach(java.lang.String, java.lang.Object)
	 */
	public void attach(String key, Object attachment) throws NoInstalledAttacher
	{
		accessProtection.lock();
		try
		{
		    IAttacher attacher = attachers.get(key);
			if (attacher == null)
				throwNoInstalledAttacher(key);
			Object object = objects.get(key);
			attacher.attach(attachment, object);   
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#attach(java.lang.String, java.lang.Object, org.almendra.commons.janet.cas.spaces.IAttacher)
	 */
	public void attach(String key, Object attachment, IAttacher attacher)
	{
		accessProtection.lock();
		try
		{
		    Object object = objects.get(key);
			attacher.attach(attachment, object);
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	/* (non-Javadoc)
	 * @see org.almendra.commons.janet.cas.spaces.IObjectSpace#detach(java.lang.String, org.almendra.commons.janet.cas.spaces.IDetacher)
	 */
	public Object detach(String key, IDetacher detacher)
	{
		accessProtection.lock();
		try
		{
		    Object object = objects.get(key);
			return detacher.detach(object); 
		}
		finally
		{
		    accessProtection.unlock();
		}
	}

	protected boolean primContains(String key)
	{
		return objects.containsKey(key);
	}
}
